functor ParseTreeFun(structure Pos: POS): PARSE_TREE =
  struct
    open Pos
    open Str
    open SympBug

    exception SympParseInternalError of string

    datatype ParseTree = 
	Program of Pos * ParseTree list
      | Include of Pos * string
      (* Included(pos,file,parse_tree) Parsed included file *)
      | Included of Pos * string * ParseTree
      | Quote of Pos * string
	(* This is for creating temporary objects in the tree and
	   replacing them later with new values, in all places at once.
	   Once all objects are final, one could walk over the tree and
	   make them permanent. *)
      | SharedRef of ParseTree ref
      | Id of Pos * string
	(* Unique ID, internal names of objects, to distinguish
	   them from the user-given names. *)
      | Uid of string
      | Tvar of Pos * string
      (* An internal type variable introduced in typechecking.
       * It means the type is not fully deduced yet; it cannot be 
       * considered as a polymorphic type variable. *)
      | TvarN of int
      (* Type of the form "<type>,...,<type> name", like "int list" *)
      | TypeInst of Pos * (ParseTree list) * ParseTree
      | Ellipsis of Pos
(*        | Module of Pos * ParseTree * (ParseTree list option) *)
(*  	              * (ParseTree option) * ParseTree *)
      (* Built-in top level declarations (like +, *, etc.) *)
      | TopClosure of ParseTree list
      | Module of Pos.Pos *
	{ name: ParseTree,
	  statparams: ParseTree list option,
	  dynparams: ParseTree option,
	  body: ParseTree
	}
      | ModuleClosure of
	{ name: ParseTree option,
	  uname: ParseTree, (* unique module identifier *)
	  dynName: ParseTree option,
	  Sig: ParseTree, (* either ModuleSig or (tp -> ModuleSig) *)
	  closure: ParseTree list,
	  (* def = NONE for begin-end, since only closure is important in
	     that case *)
	  def: ParseTree option,
	  parent: ParseTree
	}
      (* Module's "type" for internal typechecking purposes.
       * In the future it is likely to be extended to ML-like signatures. *)
      | ModuleSig of { statparams: ParseTree list option }
      (* Parent closure of an instantiated module *)
      | ModuleParamClosure of
	{ closure: ParseTree list,
	  parent: ParseTree
	}
      (* This is passed as a "parent closure" to inner functions.
	 Originally it contains "FunClosure" that only defines
	 the name space, and later it's overwritten by the actual
	 function definition, thus making a circular structure for
	 recursive functions *)
(*        | FunClosureRef of ParseTree ref *)
      (* This is a preliminary version of the  "parent closure",
       * without the actual definition.  It is used only at 
       * the time of typechecking, and is later replaced by the full
       * version. Maybe, it's not even needed... *)
(*        | FunClosureSmall of *)
(*  	{ name: ParseTree option, *)
(*  	  formals: ParseTree, (* There can only be one argument *) *)
(*  	  parent: ParseTree (* The parent's closure *) *)
(*  	} *)
      (* The full version of the above.  If body = Fake, then it's the
       * same as the above. *)
      | FunClosure of
	{ name: ParseTree option,
	  formals: ParseTree, (* There can only be one argument *)
	  parent: ParseTree, (* The parent's closure *)
	  body: ParseTree
	}
      (* The "linearized" representation of recursive function `f':
       * RecurFun(fn f => <the curried def. of `f'>) *)
      | RecurFun of ParseTree
      (* Closure for `let' expression *)
      | LetClosure of
	{ locals: ParseTree list,
	  parent: ParseTree,
	  body: ParseTree }
      (* Same for assignments *)
      | LetAsstClosure of
	{ locals: ParseTree list,
	  parent: ParseTree,
	  body: ParseTree list }
      (* Serves as an additional "context" for recursive or
       * parameterized type declarations. *)
      | TypeClosure of { name: ParseTree,
			 uname: ParseTree,
			 params: ParseTree list,
			 def: ParseTree,
			 recursive: bool,
			 parent: ParseTree }
      (* Keeps bound variables in a choice of a `case' statement *)
      | ChoiceClosure of { pattern: ParseTree,
			   (* Internal name for selector expr *)
			   uname: ParseTree,
			   names: ParseTree list,
			   body: ParseTree,
			   parent: ParseTree }
      (* Same for assignments *)
      | ChoiceAsstClosure of
	{ pattern: ParseTree,
	  (* Internal name for selector expr *)
	  uname: ParseTree,
	  names: ParseTree list,
	  body: ParseTree list,
	  parent: ParseTree }

      (* The closure for the CHOOSE operator.  It's the same as ChoiceClosure,
       * only serves a different purpose. *)
      | ChooseClosure of { pattern: ParseTree,
			   Type: ParseTree, (* Type of the result *)
			   names: ParseTree list,
			   body: ParseTree,
			   parent: ParseTree }
      (* Quantifiers *)
      | ForallClosure of { names: ParseTree list,
			   body: ParseTree,
			   parent: ParseTree }
      | ExistsClosure of { names: ParseTree list,
			   body: ParseTree,
			   parent: ParseTree }
      | QuantifiedVar of { name: ParseTree,
			   uname: ParseTree,
			   Type: ParseTree }
      (* Module parallel composition *)
      | SyncClosure of { names: ParseTree list,
			 body: ParseTree,
			 parent: ParseTree }
      | AsyncClosure of { names: ParseTree list,
			  body: ParseTree,
			  parent: ParseTree }

      (* We probably do not need this any longer *)
      (**************************************************************)

      (* Object's definition can be either some expression, or: *)
      (* This is a builtin top-level object *)
      | Builtin of {name: ParseTree, Type: ParseTree}
      (* A state variable object *)
      | StateVar of
	{ name: ParseTree,
	  uname: ParseTree,
	  Type: ParseTree,
	  (* State variables have to be distinguished later for
	     different module instances.  The `id' field will hold the
	     string of module instances (e.g. M1[i].M2[j].M3(y,x)) the
	     state var is coming from, where Mi are "uname"'s of
	     modules.  The context provides the direct link to the
	     transition system information. *)
	  id: ParseTree
	}
      (* Object is a typechecked named object representation.
         It must carry enough info to disambiguate overloaded
	 identifiers. *)
      | Object of
	{ name: ParseTree,
	  uname: ParseTree,
	  Type: ParseTree,
	  def: ParseTree
	}
      (* Object instance that replaces a name in the program.  The
	 `subst' field is a function that properly instantiates the
         types within the object for this particular instance. *)
      | ObjectInst of {obj: ParseTree,
		       subst: ParseTree -> ParseTree}

      (**************************************************************)

      (* A variable bound by a pattern *)
      | PatternFormal of { name: ParseTree,
			   uname: ParseTree,
			   Type: ParseTree,
			   (* Function that wraps a proper extracter around 
			    * an actual parameter *)
			   extract: ParseTree -> ParseTree }
      (* A variable bound by a pattern in dynamic parameter of a module *)
      | DynPatternFormal of { name: ParseTree,
			      uname: ParseTree,
			      Type: ParseTree,
			     (* Function that wraps a proper extracter around 
			      * an actual parameter *)
			      extract: ParseTree -> ParseTree }
      (* Datatype constructor. 
       * The type is either the datatype itself (EnumType), or
       * paramType -> datatype (FunType). *)
      | TypeConstr of { name: ParseTree,
		        uname: ParseTree,
		        Type: ParseTree }
      (* Record field object.
         `uname' doesn't really make sense here. *)
      | RecordField of { name: ParseTree,
			 Type: ParseTree
		       (* Avoid circularity.  We probably do not need it. *)
		       (* recordType: ParseTree *) }
      (* Static formal constant and type parameters of a module.
         Type parameters are represented by PType. *)
      | StaticFormalConst of
	{ name: ParseTree,
	  uname: ParseTree,
	  Type: ParseTree }
      | StaticFormalType of
	{ name: ParseTree,
	  uname: ParseTree }
      | Theorem of Pos * ParseTree * ParseTree
      (* Parametric type in the static params of a module *)
      | PType of Pos * ParseTree
      | Symmetric of Pos * ParseTree
      | Finite of Pos * ParseTree
      | BeginEnd of Pos * ParseTree list
      | Sync2 of Pos * ParseTree * ParseTree
      | Async2 of Pos * ParseTree * ParseTree
      | Sync of Pos * ParseTree list * ParseTree
      | Async of Pos * ParseTree list * ParseTree
      (* TypeDecl(pos,name,paramlist,def) *)
      | TypeDecl of Pos * ParseTree * (ParseTree list) * ParseTree
      | DatatypeDecl of Pos * ParseTree * (ParseTree list) * ParseTree
      | Var of Pos * ParseTree list
      | StateVarDecl of Pos * ParseTree list
      | Val of Pos * ParseTree * ParseTree
      | FunCases of Pos * ParseTree list
      (* FunCase(pos,name,patterns,result_type,expr) *)
      | FunCase of Pos * ParseTree * (ParseTree list)
	* (ParseTree option) * ParseTree
      | Open of Pos * ParseTree
      | Export of Pos * ParseTree list
      | Models of Pos * ParseTree * ParseTree
      | Refines of Pos * ParseTree * ParseTree
      | Nop of Pos
      | NormalAsst of Pos * ParseTree * ParseTree
      | InitAsst of Pos * ParseTree * ParseTree
      | NextAsst of Pos * ParseTree * ParseTree
      (* Wrapper to represent next(x) in boolean formulas for the
         transition relation (internal use only) *)
      | Next of ParseTree
      | IfAsst of Pos * (ParseTree list) * ParseTree list
      | CondAsst of Pos * ParseTree * (ParseTree list)
      | LetAsst of Pos * (ParseTree list) * (ParseTree list)
      | CaseAsst of Pos * ParseTree * (ParseTree list)
      | ChoiceAsst of Pos * ParseTree * (ParseTree list)
      | ChooseAsst of Pos * ParseTree list
      | IfExpr of Pos * (ParseTree list) * ParseTree
      | CondExpr of Pos * ParseTree * ParseTree
      | LetExpr of Pos * (ParseTree list) * ParseTree
      | CaseExpr of Pos * ParseTree * (ParseTree list)
      | ChoiceExpr of Pos * ParseTree * ParseTree
      | ChooseExpr of Pos * ParseTree * ParseTree
      | WithExpr of Pos * ParseTree * (ParseTree list)
      | WithAsst of Pos * ParseTree * ParseTree
      | RecordExpr of Pos * ParseTree list
      | TupleExpr of Pos * ParseTree list
      | NondetExpr of Pos * ParseTree list
      | RecordAsst of Pos * ParseTree * ParseTree
      (* ModuleInst(name,stat_param_list).
       Dynamic params are applied with Appl. *)
      | ModuleInst of Pos * ParseTree * (ParseTree list)
      | Dot of Pos * ParseTree * ParseTree
      | Type of Pos * ParseTree
      | Datatype of Pos * ParseTree
      | And of Pos
      | Or of Pos
      | Implies of Pos
      (* Normally, these two will also be "Implies", but the user may 
       * redefine them to something else *)
      | Arrow of Pos
      | Darrow of Pos
      | Iff of Pos
      | Eq of Pos
      | NotEq of Pos
      | Not of Pos
      | Forall of Pos * (ParseTree list) * ParseTree
      | Exists of Pos * (ParseTree list) * ParseTree
      | Ag of Pos
      | Af of Pos
      | Eg of Pos
      | Ef of Pos
      | Ax of Pos
      | Ex of Pos
      | Au of Pos
      | Eu of Pos
      | Ar of Pos
      | Er of Pos
      | Lt of Pos
      | Gt of Pos
      | Le of Pos
      | Ge of Pos
      | Plus of Pos
      | Minus of Pos
      | UMinus of Pos
      | Times of Pos
      | Div of Pos
      | Mod of Pos
      | Number of Pos * int
      | BoolType of Pos
      (* Internal type for 2-valued boolean operations *)
      | BoolType2
      | NatType of Pos
      | IntType of Pos
      | RangeType of Pos * ParseTree * ParseTree
      | FunType of Pos * ParseTree * ParseTree
      | TupleType of Pos * ParseTree list
      | ArrayType of Pos * ParseTree * ParseTree
      | RecordType of Pos * ParseTree list
      | EnumType of Pos * ParseTree list
      | AbstractType of ParseTree list
      | Of of Pos * ParseTree * ParseTree
      | TypedVars of Pos * (ParseTree list) * (ParseTree option)
      | False of Pos
      | True of Pos
      (* Special 2-valued true and false, for the internal use of the
         transition relation generator *)
      | False2
      | True2
      | Min of Pos
      | Max of Pos
      | Undefined of Pos
      | Anyvalue of Pos
      (* Internal value representing "all the other values" in abstractions *)
      | OtherValue
      | Self of Pos
      | Appl of Pos * ParseTree * ParseTree
      | TypedExpr of Pos * ParseTree * ParseTree
      | Fn of Pos * ParseTree list
      | ChoicePattern of Pos * ParseTree list
      | ApplPattern of Pos * ParseTree * ParseTree
      | TypedPattern of Pos * ParseTree * ParseTree
      | TuplePattern of Pos * ParseTree list
      | RecordPattern of Pos * ParseTree list
	(* AsPattern(p,id,pattern) represents `id as pattern' *)
      | AsPattern of Pos * ParseTree * ParseTree
      | Underscore of Pos
      (* And pattern extractors, for internal use only, to provide a way
       * of assigning values to variables in the pattern *)
      (* ExtractAppl(DatatypeConstructer,Expr) *)
      | ExtractAppl of ParseTree * ParseTree
      (* Takes n-th element of a tuple *)
      | ExtractTuple of int * ParseTree
      (* ExtractRecord(FieldName, Expr) - no full record type is needed,
       * the expression must contain the field info *)
      | ExtractRecord of ParseTree * ParseTree
      (* ExtractIndex(Expr) - get the TypeConstr that makes the value
         of `Expr' that is of EnumType *)
      | ExtractIndex of ParseTree

      (* for emphasizing grouping, e.g. for tuple type *)
      | Group of Pos * ParseTree 
      | Fake (* for some errors and non-implemented commands *)

    fun isInfix(And _) = true
      | isInfix(Or _) = true
      | isInfix(Implies _) = true
      | isInfix(Arrow _) = true
      | isInfix(Darrow _) = true
      | isInfix(Iff _) = true
      | isInfix(Eq _) = true
      | isInfix(NotEq _) = true
      | isInfix(Plus _) = true
      | isInfix(Minus _) = true
      | isInfix(Times _) = true
      | isInfix(Div _) = true
      | isInfix(Mod _) = true
      | isInfix(Object{def=Builtin{name=n,...},...}) = isInfix n
      | isInfix(ObjectInst{obj=obj,...}) = isInfix obj
      | isInfix(Builtin{name=n,...}) = isInfix n
      | isInfix _ = false

    fun isUnary(Not _) = true
      | isUnary(UMinus _) = true
      | isUnary(Ag _) = true
      | isUnary(Af _) = true
      | isUnary(Eg _) = true
      | isUnary(Ef _) = true
      | isUnary(Ax _) = true
      | isUnary(Ex _) = true
      | isUnary(Object{def=Builtin{name=n,...},...}) = isUnary n
      | isUnary(ObjectInst{obj=obj,...}) = isUnary obj
      | isUnary _ = false
	
    fun pt2strDebug tree =
	let val op * = fn (x,y) => Conc(x,y)
	    fun list2str ff sep (x::y::tl) = 
		(loop ff x)*(Str sep)*(list2str ff sep (y::tl))
	      | list2str ff _ ([x]) = loop ff x
	      | list2str _ _ [] = Str ""
	    and option2str _ NONE = Str ""
	      | option2str ff (SOME tree) = loop ff tree
	    and optionLst2str _ _ NONE = Str ""
	      | optionLst2str ff sep (SOME lst) = list2str ff sep lst
	    and loop ff (Program(_,lst)) = (list2str ff ";\n" lst)
	      | loop ff (Include(_,str)) = (Str("include(\""^str^"\")"))
	      | loop ff (Included(_,str,x)) = 
		(Str("-- include(\""^str^"\")\n"))*(loop ff x)
		*(Str("\n-- end of file "^str^"\n"))
	      | loop ff (Quote(_,str)) = (Str ("\""^str^"\""))
	      | loop ff (SharedRef x) = (Str "SharedRef(")*(loop ff(!x))*(Str ")")
	      | loop ff (Id(_,str)) = Str str
	      | loop ff (Uid(str)) = Str("<"^str^">")
	      | loop ff (Tvar(_,str)) = Str("'"^str)
	      | loop ff (TvarN n) = Str("'X!"^(Int.toString n))
	      | loop ff (TypeInst(_,parms,tp)) =
		  (Str "(")*(list2str ff "," parms)*(Str ") ")
		  *(loop ff (ff tp))
	      | loop ff (Ellipsis _) = Str "..."
	      | loop ff (TopClosure lst) =
		(Str "\n -- Top Closure\n  ")
		*(list2str ff ";\n  " (List.rev lst))
		*(Str "\n -- End of Top Closure\n")
	      | loop ff (Module(_,{ name=name,
				   statparams=stat,
				   dynparams=dyn,
				   body=body})) = 
		(Str "module ")*(loop ff name)
		  *(case stat of
			NONE => Str ""
		      | SOME lst => (Str "[")*(list2str ff "," lst)*(Str "]"))
		  *(case dyn of
			NONE => Str ""
		      | SOME p => loop ff p)
		  *(Str " = ")*(loop ff body)
	      | loop ff (ModuleClosure{name=name,
				       uname=uname,
				       dynName=dn,
				       Sig=ms,
				       def=def,
				       closure=cl,
				       parent=p}) =
		  (Str "ModuleClosure{name=")
		  *(case name of
			NONE => (Str "<unnamed>")
		      | SOME n => (loop ff n))
		  *(Str ",\n              uname=")
		  *(loop ff uname)
		  *(Str ",\n              dynName=")
		  *(case dn of
			NONE => (Str "<unnamed>")
		      | SOME n => (loop ff n))
		  *(Str ",\n              Sig=")*(loop ff ms)
		  *(Str ",\n              closure=[")
		  *(list2str ff ",\n                 " (List.rev cl))
		  *(Str "],\n              def=")
		  *(case def of
			NONE => (Str "NONE")
		      | SOME d => (loop ff d))
		  *(Str ",\n              parent=")
		  *(loop ff p) (*  *(Str "<skipping...>") *)
		  *(Str "}\n")
	      | loop ff (ModuleSig{statparams=stat}) =
		  (Str "ModuleSig{statparams=")
		  *(case stat of
			NONE => Str "[]"
		      | SOME lst => (Str "[")*(list2str ff "," lst)*(Str "]"))
		  *(Str "}")
	      | loop ff (ModuleParamClosure{closure=cl, parent=p}) =
		  (Str "ModuleParamClosure{closure = [")
		  *(list2str ff ", " (List.rev cl))
		  *(Str "],\n                 parent = ")
		  (* *(loop ff p) *) *(Str "<skipping...>")
		  *(Str "}\n")
	      | loop ff (FunClosure{ name=name,
				 formals=parms,
				 body=body, ...}) =
		  (Str "FunClosure{ name=")
		  *(case name of
			NONE => (Str "<unnamed>")
		      | SOME(n) => (loop ff n))
		  *(Str ", formals=")*(loop ff parms)
		  *(Str ", body=")*(loop ff body)
		  *(Str "}")
	      | loop ff (RecurFun f) = (Str "RecurFun(")*(loop ff f)*(Str ")")
	      | loop ff (LetClosure{locals=locs,parent=p,body=b}) =
		  (Str "LetClosure{locals=[")*(list2str ff ", " (List.rev locs))
		  *(Str "],\n           parent=")(* *(loop ff p) *)*(Str "<skipping...>")
		  *(Str ",\n           body=")*(loop ff b)
		  *(Str "}\n")
	      | loop ff (LetAsstClosure{locals=locs,parent=p,body=b}) =
		  (Str "LetAsstClosure{locals=[")*(list2str ff ", " (List.rev locs))
		  *(Str "],\n           parent=")(* *(loop ff p) *)*(Str "<skipping...>")
		  *(Str ",\n           body=[")*(list2str ff ";\n    " b)
		  *(Str "] }\n")
	      | loop ff (TypeClosure{name=name,uname=uname,params=parms,
				     def=def,recursive=r,parent=p}) =
		  (* Reset the substitution - the type closure has no
		     external type vars. *)
		  let val ff = (fn x=>x)
		  in (Str "TypeClosure{name=")*(loop ff name)
		      *(Str ",\n            uname=")*(loop ff uname)
		      *(Str ",\n            params=[")*(list2str ff ", " parms)
		      *(Str "],\n")
		      *(Str "            def=")*(loop ff def)*(Str ",\n")
		      *(if r then (Str "            recursive=true,\n")
			else (Str "            recursive=false,\n"))
		      *(Str "            parent=")(* *(loop ff p) *)
		      *(Str "<skipping...>")
		      *(Str "}\n")
		  end
	      | loop ff (ChoiceClosure{pattern=pat,
				       uname=uname,
				       names=names,body=b,parent=p}) =
		  (Str "ChoiceClosure{pattern=")*(loop ff pat)*(Str ",\n")
		  *(Str "              uname=")*(loop ff uname)*(Str ",\n")
		  *(Str "              names=[")
		  *(list2str ff "," names)*(Str "],\n")
		  *(Str "              body=")*(loop ff b)*(Str ",\n")
		  *(Str "              parent=")(* *(loop ff p) *)*(Str "<skipping...>")
		  *(Str "}\n")
	      | loop ff (ChoiceAsstClosure{pattern=pat,
					   uname=uname,
					   names=names,
					   body=b,parent=p}) =
		  (Str "ChoiceClosure{pattern=")*(loop ff pat)*(Str ",\n")
		  *(Str "              uname=")*(loop ff uname)*(Str ",\n")
		  *(Str "              names=[")
		  *(list2str ff "," names)*(Str "],\n")
		  *(Str "              body=[\n   ")
		  *(list2str ff "\n   " b)*(Str "],\n")
		  *(Str "              parent=")(* *(loop ff p) *)*(Str "<skipping...>")
		  *(Str "}\n")
	      | loop ff (ChooseClosure{pattern=pat,Type=tp,names=names,body=b,parent=p}) =
		  (Str "ChooseClosure{pattern=")*(loop ff pat)*(Str ",\n")
		  *(Str "              Type=")*(loop ff (ff tp))*(Str ",\n")
		  *(Str "              names=[")
		  *(list2str ff "," (List.rev names))*(Str "],\n")
		  *(Str "              body=")*(loop ff b)*(Str ",\n")
		  *(Str "              parent=")(* *(loop ff p) *)*(Str "<skipping...>")
		  *(Str "}\n")
	      | loop ff (ForallClosure{names=names,body=body,parent=p}) =
		  (Str "ForallClosure{names=[")*(list2str ff "," names)
		  *(Str "],\n              body=")*(loop ff body)
		  *(Str ",\n              parent=")(* *(loop ff p) *)
		  *(Str "<skipping...>")*(Str "}\n")
	      | loop ff (ExistsClosure{names=names,body=body,parent=p}) =
		  (Str "ExistsClosure{names=[")*(list2str ff "," names)
		  *(Str "],\n              body=")*(loop ff body)
		  *(Str ",\n              parent=")(* *(loop ff p) *)
		  *(Str "<skipping...>")*(Str "}\n")
	      | loop ff (QuantifiedVar{name=n,uname=uname,Type=tp}) =
		  (Str "QuantifiedVar{name=")*(loop ff n)
		  *(Str ",\n              uname=")*(loop ff uname)
		  *(Str ",\n              Type=")*(loop ff (ff tp))
		  *(Str " }")
	      | loop ff (SyncClosure{names=names,body=body,parent=p}) =
		  (Str "SyncClosure{names=[")*(list2str ff "," names)
		  *(Str "],\n            body=")*(loop ff body)
		  *(Str ",\n            parent=")(* *(loop ff p) *)
		  *(Str "<skipping...>")*(Str "}\n")
	      | loop ff (AsyncClosure{names=names,body=body,parent=p}) =
		  (Str "AsyncClosure{names=[")*(list2str ff "," names)
		  *(Str "],\n             body=")*(loop ff body)
		  *(Str ",\n             parent=")(* *(loop ff p) *)
		  *(Str "<skipping...>")*(Str "}\n")
	      | loop ff  (Builtin{name=n,Type=tp}) =
		  (Str "<Builtin(")*(loop ff n)*(Str ": ")*(loop ff (ff tp))*(Str ")>")
	      | loop ff (StateVar{name=name,uname=uname,Type=tp,id=id}) =
		  (Str "StateVar{name=")*(loop ff name)
		  *(Str ",\n         uname=")*(loop ff uname)
		  *(Str ",\n         Type=")*(loop ff (ff tp))
		  *(Str ",\n         id=")*(loop ff id)*(Str "}")
	      | loop ff (Object{name=name,uname=uname,Type=tp,def=def}) =
		  (Str "{{")*(loop ff name)*(Str " ")*(loop ff uname)
		  *(Str " : ")*(loop ff (ff tp))
		  *(Str " = ")*(loop ff def)*(Str "}}")
	      (* Here we take the new substitution from ObjectInst *) 
	      | loop ff (ObjectInst{obj=obj,subst=ff1}) =
(*  		  (Str "@")*(loop ff obj) *)
		  (Str "@(")*(loop (ff o ff1) obj)*(Str ")")
	      | loop ff (PatternFormal{name=name,uname=uname,Type=tp,...}) =
		  (Str "<[")*(loop ff name)*(Str " ")*(loop ff uname)
		  *(Str " : ")*(loop ff (ff tp))
		  *(Str "]>")
	      | loop ff (DynPatternFormal{name=name,uname=uname,Type=tp,...}) =
		  (Str "#<[")*(loop ff name)*(Str " ")*(loop ff uname)
		  *(Str " : ")*(loop ff (ff tp))
		  *(Str "]>")
	      | loop ff (TypeConstr{name=name,uname=uname,Type=tp}) =
		  (Str "[<")*(loop ff name)*(Str " ")*(loop ff uname)
		  *(Str " : ")*(loop ff (ff tp))*(Str ">]")
	      | loop ff (RecordField{name=name,Type=tp}) =
		  (Str "<{")*(loop ff name)*(Str " : ")*(loop ff (ff tp))
		  *(Str "}>")
	      | loop ff (StaticFormalConst{name=name,uname=uname,Type=tp}) =
		  (Str "[[")*(loop ff name)*(Str " ")*(loop ff uname)
		  *(Str " : ")*(loop ff (ff tp))*(Str "]]")
	      | loop ff (StaticFormalType{name=name,uname=uname}) =
		  (Str "[[")*(loop ff name)*(Str " ")*(loop ff uname)
		  *(Str " : type]]")
	      | loop ff (Theorem(_,name,body)) = 
		  (Str "theorem ")*(loop ff name)*(Str " = ")*(loop ff body)
	      | loop ff (PType(_,tp)) = (Str "type ")*(loop ff (ff tp))
	      | loop ff (Symmetric(_,tp)) = 
		  (Str "symmetric ")*(loop ff (ff tp))
	      | loop ff (Finite(_,tp)) = (Str "finite ")*(loop ff (ff tp))
	      | loop ff (BeginEnd(_,exprs)) = 
		  (Str "\nbegin\n  ")
		  *(list2str ff ";\n  " exprs)
		  *(Str "\nend -- module\n")
	      | loop ff (Sync2(_,e1,e2)) = 
		  (Str "(")*(loop ff e1)*(Str ")")
		  *(Str " || ")*(Str "(")*(loop ff e2)*(Str ")")
	      | loop ff (Async2(_,e1,e2)) = 
		  (Str "(")*(loop ff e1)*(Str ")")
		  *(Str " | ")*(Str "(")*(loop ff e2)*(Str ")")
	      | loop ff (Sync(_,vars,e)) =
		  (Str "sync ")*(list2str ff ", " vars)*(Str ": ")*(loop ff e)
	      | loop ff (Async(_,vars,e)) = 
		  (Str "async ")*(list2str ff ", " vars)*(Str ": ")*(loop ff e)
	      | loop ff (TypeDecl(_,name,parms,tp)) = 
		  (Str "type ")
		  *(case parms of
		       [] => (Str "")
		     | _ => (Str "(")*(list2str ff "," parms)*(Str ") "))
		  *(loop ff name)*(Str " = ")*(loop ff (ff tp))
	      | loop ff (DatatypeDecl(_,name,parms,tp)) = 
		  (Str "datatype ")
		  *(case parms of
		       [] => (Str "")
		     | _ => (Str "(")*(list2str ff "," parms)*(Str ") "))
		  *(loop ff name)
		  *(Str " = ")*(loop ff (ff tp))
	      | loop ff (Var(_,vars)) = 
		  (Str "var ")*(list2str ff ", " vars)
	      | loop ff (StateVarDecl(_,vars)) = 
		  (Str "statevar ")*(list2str ff ", " vars)
	      | loop ff (Val(_,name,e)) = 
		  (Str "val ")*(loop ff name)*(Str " = ")*(loop ff e)
	      | loop ff (FunCases(_,lst)) = 
		  (Str "fun ")*(list2str ff "\n  | " lst)
	      | loop ff (FunCase(_,name,patterns,res_tp,e)) = 
		  (loop ff name)*(Str " ")*(list2str ff " " patterns)
		  *(case res_tp of
			NONE => Str ""
		      | SOME tp => (Str " : ")*(loop ff (ff tp)))
		  *(Str " = ")*(loop ff e)
	      | loop ff (Open(_,e)) = (Str "open ")*(loop ff e)
	      | loop ff (Export(_,lst)) = 
		  (Str "export ")
		  *(case lst of
			[] => (Str "all")
		      | _  => list2str ff ", " lst)
	      | loop ff (Models(_,e1,e2)) = 
		  (Str "(")*(loop ff e1)*(Str ") |= (")*(loop ff e2)*(Str ")")
	      | loop ff (Refines(_,e1,e2)) = 
		  (Str "(")*(loop ff e1)*(Str ") refines (")*(loop ff e2)*(Str ")")
	      | loop ff (Nop _) = Str "nop"
	      | loop ff (NormalAsst(_,name,e)) = 
		  (Str "(")*(loop ff name)*(Str ") := ")*(loop ff e)
	      | loop ff (InitAsst(_,name,e)) = 
		  (Str "init(")*(loop ff name)*(Str ") := ")*(loop ff e)
	      | loop ff (NextAsst(_,name,e)) = 
		  (Str "next(")*(loop ff name)*(Str ") := ")*(loop ff e)
	      | loop ff (Next e) = (Str "next(")*(loop ff e)*(Str ")")
	      | loop ff (IfAsst(_,cond_lst,else_lst)) = 
		  (Str "if ")
		  *(list2str ff "\nelsif\n  " cond_lst)
		  *(Str "\nelse\n  ")*(list2str ff "\n  " else_lst)
		  *(Str "\nendif ")
	      | loop ff (CondAsst(_,e,assts)) = 
		  (loop ff e)*(Str " then\n  ")*(list2str ff "\n  " assts)
	      | loop ff (LetAsst(_,locals,assts)) = 
		  (Str "let ")*(list2str ff "\n    " locals)
		  *(Str "\nin\n  ")*(list2str ff ";\n  " assts)
		  *(Str "\nend ")
	      | loop ff (CaseAsst(_,e,cases)) = 
		  (Str "case ")*(loop ff e)*(Str " of\n    ")
		  *(list2str ff "\n  | " cases)
		  *(Str "\nendcase ")
	      | loop ff (ChoiceAsst(_,e,assts)) =
		  (loop ff e)
		  *(Str " =>\n      ")
		  *(list2str ff ";\n      " assts)
	      | loop ff (ChooseAsst(_,lst)) = 
		  (Str "choose\n    ")
		  *(list2str ff "\n  | " lst)
		  *(Str "\nendchoose ")
	      | loop ff (IfExpr(_,conds,else_exp)) = 
		  (Str "if ")*(list2str ff "\nelsif " conds)
		  *(Str "\nelse ")*(loop ff else_exp)
		  *(Str "\nendif ")
	      | loop ff (CondExpr(_,e1,e2)) = 
		  (loop ff e1)*(Str " then ")*(loop ff e2)
	      | loop ff (LetExpr(_,locals,e)) = 
		  (Str "let ")
		  *(list2str ff "\n    " locals)
		  *(Str "\nin ")*(loop ff e)*(Str "\nend ")
	      | loop ff (CaseExpr(_,e,cases)) = 
		  (Str "case ")*(loop ff e)*(Str " of\n    ")
		  *(list2str ff "\n  | " cases)
		  *(Str "\nendcase ")
	      | loop ff (ChoiceExpr(_,p,e)) = 
		  (loop ff p)*(Str " => ")*(loop ff e)
	      | loop ff (ChooseExpr(_,p,e)) = 
		  (Str "pick ")*(loop ff p)*(Str ": ")*(loop ff e)
	      | loop ff (WithExpr(_,e,lst)) = 
		  (loop ff e)*(Str " with [ ")*(list2str ff ", " lst)*(Str " ]")
	      | loop ff (WithAsst(_,name,e)) = 
		  (loop ff name)*(Str " = ")*(loop ff e)
	      | loop ff (RecordExpr(_,lst)) = 
		  (Str "{ ")*(list2str ff ", " lst)*(Str " }")
	      | loop ff (TupleExpr(_,lst)) = 
		  (Str "( ")*(list2str ff ", " lst)*(Str " )")
	      | loop ff (NondetExpr(_,lst)) = 
		  (Str "(")*(list2str ff ") | (" lst)*(Str ")")
	      | loop ff (RecordAsst(_,name,e)) = 
		  (loop ff name)*(Str " = ")*(loop ff e)
	      | loop ff (ModuleInst(_,name,stats)) = 
		  (loop ff name)*(Str "[")*(list2str ff ", " stats)*(Str "]")
	      | loop ff (Dot(_,e1,e2)) = 
		  (loop ff e1)*(Str ".")*(loop ff e2)
	      | loop ff (Type(_,e)) = 
		  (Str "type ")*(loop ff e)
	      | loop ff (Datatype(_,e)) = 
		  (Str "\ndatatype ")*(loop ff e)
	      | loop ff (And _) = (Str "and")
	      | loop ff (Or _) = (Str "or")
	      | loop ff (Implies _) = (Str "implies")
	      | loop ff (Arrow _) = (Str "->")
	      | loop ff (Darrow _) = (Str "=>")
	      | loop ff (Iff _) = (Str "<=>")
	      | loop ff (Eq _) = (Str "=")
	      | loop ff (NotEq _) = (Str "!=")
	      | loop ff (Not _) = (Str "!")
	      | loop ff (Forall(_,vars,e)) = 
		  (Str "forall ")*(list2str ff ", " vars)*(Str ": ")*(loop ff e)
	      | loop ff (Exists(_,vars,e)) = 
		  (Str "exists ")*(list2str ff ", " vars)*(Str ": ")*(loop ff e)
	      | loop ff (Ag _) = (Str "AG")
	      | loop ff (Af _) = (Str "AF")
	      | loop ff (Eg _) = (Str "EG")
	      | loop ff (Ef _) = (Str "EF")
	      | loop ff (Ax _) = (Str "AX")
	      | loop ff (Ex _) = (Str "EX")
	      | loop ff (Au _) = (Str "AU")
	      | loop ff (Eu _) = (Str "EU")
	      | loop ff (Ar _) = (Str "AR")
	      | loop ff (Er _) = (Str "ER")
	      | loop ff (Lt _) = (Str "<")
	      | loop ff (Gt _) = (Str ">")
	      | loop ff (Le _) = (Str "<=")
	      | loop ff (Ge _) = (Str ">=")
	      | loop ff (Plus _) = (Str "+")
	      | loop ff (Minus _) = (Str "-")
	      | loop ff (UMinus _) = (Str "~")
	      | loop ff (Times _) = (Str "*")
	      | loop ff (Div _) = (Str "/")
	      | loop ff (Mod _) = (Str "mod")
	      | loop ff (Number(_,n)) = (Str (Int.toString n))
	      | loop ff (BoolType _) = (Str "bool")
	      | loop ff (BoolType2) = (Str "bool2")
	      | loop ff (NatType _) = (Str "nat")
	      | loop ff (IntType _) = (Str "int")
	      | loop ff (RangeType(_,x,y)) =
		  (Str "[")*(loop ff x)*(Str "..")*(loop ff y)*(Str "]")
	      | loop ff (FunType(_,tp1,tp2)) =
		  (Str "(")*(loop ff (ff tp1))*(Str ")")
		  *(Str " -> ")*(Str "(")*(loop ff (ff tp2))*(Str ")")
	      | loop ff (TupleType(_,lst)) = 
		  (Str "(")*(list2str ff ") * (" lst)*(Str ")")
	      | loop ff (ArrayType(_,tp1,tp2)) = 
		  (Str "array ")*(loop ff (ff tp1))*(Str " of ")
		  *(loop ff (ff tp2))
	      | loop ff (RecordType(_,lst)) = 
		  (Str "{ ")*(list2str ff ", " lst)*(Str " }")
	      | loop ff (EnumType(_,lst)) = 
		  (Str "(")*(list2str ff " | " lst)*(Str ")")
	      | loop ff (AbstractType(lst)) = 
		  (Str "AbstractType[")*(list2str ff ", " lst)*(Str "]")
	      | loop ff (Of(_,name,tp)) = 
		  (loop ff name)*(Str " of ")*(loop ff (ff tp))
	      | loop ff (TypedVars(_,vars,tp)) = 
		  (Str "(")*(list2str ff ", " vars)
		  *(case tp of
			NONE => (Str "")
		      | SOME tp => (Str ": ")*(loop ff (ff tp)))
		  *(Str ")")
	      | loop ff (False _) = (Str "false")
	      | loop ff (True _) = (Str "true")
	      | loop ff (False2) = (Str "false2")
	      | loop ff (True2) = (Str "true2")
	      | loop ff (Min _) = (Str "min")
	      | loop ff (Max _) = (Str "max")
	      | loop ff (Undefined _) = (Str "undefined")
	      | loop ff (Anyvalue _) = (Str "anyvalue")
	      | loop ff OtherValue = (Str "OtherValue")
	      | loop ff (Self _) = (Str "self")
	      | loop ff (Appl(_,Au _,TupleExpr(_,[e1,e2]))) =
		  (Str "Always[")*(loop ff e1)*(Str " Until ")*(loop ff e2)*(Str "]")
	      | loop ff (Appl(_,Eu _,TupleExpr(_,[e1,e2]))) =
		  (Str "Sometimes[")*(loop ff e1)*(Str " Until ")
		  *(loop ff e2)*(Str "]")
	      | loop ff (Appl(_,Ar _,TupleExpr(_,[e1,e2]))) =
		  (Str "Always[")*(loop ff e1)*(Str " Releases ")
		  *(loop ff e2)*(Str "]")
	      | loop ff (Appl(_,Er _,TupleExpr(_,[e1,e2]))) =
		  (Str "Sometimes[")*(loop ff e1)*(Str " Releases ")
		  *(loop ff e2)*(Str "]")
	      | loop ff (Appl(_,e1,e2)) = 
		  if isInfix(e1) then
		      (case e2 of
			   TupleExpr(_,[x,y]) =>
			       (Str "(")*(loop ff x)*(Str " ")
			       *(loop ff e1)*(Str " ")*(loop ff y)*(Str ")")
			 | x => (Str "(op ")*(loop ff e1)*(Str " ")
			       *(loop ff e2)*(Str ")"))
		  else if isUnary(e1) then
		      (loop ff e1)*(Str "(")*(loop ff e2)*(Str ")")
		  else
		      (Str "(")*(loop ff e1)*(Str " ")*(loop ff e2)*(Str ")")
	      | loop ff (TypedExpr(_,e,tp)) = 
		  (Str "((")*(loop ff e)*(Str "): ")
		  *(loop ff (ff tp))*(Str ")")
	      | loop ff (Fn(_,lst)) =
		  (Str "fn ")*(list2str ff "\n  | " lst)
	      | loop ff (ChoicePattern(_,lst)) = (list2str ff " | " lst)
	      | loop ff (ApplPattern(_,p1,p2)) =
		  (Str "(")*(loop ff p1)*(Str " ")*(loop ff p2)*(Str ")")
	      | loop ff (AsPattern(_,p1,p2)) =
		  (Str "(")*(loop ff p1)*(Str " as ")*(loop ff p2)*(Str ")")
	      | loop ff (TypedPattern(_,p,tp)) = 
		  (Str "(")*(loop ff p)*(Str ": ")*(loop ff (ff tp))*(Str ")")
	      | loop ff (TuplePattern(_,lst)) = 
		  (Str "(")*(list2str ff ", " lst)*(Str ")")
	      | loop ff (RecordPattern(_,lst)) = 
		  (Str "{")*(list2str ff ", " lst)*(Str "}")
	      | loop ff (Underscore _) = (Str "_")
	      | loop ff (ExtractAppl(c,e)) =
		  (Str "ExtractAppl(")*(list2str ff ", " [c,e])
		  *(Str ")")
	      | loop ff (ExtractTuple(n,e)) =
		  (Str("ExtractTuple("^(Int.toString n)^","))
		  *(loop ff e)*(Str ")")
	      | loop ff (ExtractRecord(name,e)) =
		  (Str "ExtractRecord(")*(loop ff name)*(Str ",")
		  *(loop ff e)*(Str ")")
	      | loop ff (ExtractIndex e) =
		  (Str "ExtractIndex(")*(loop ff e)*(Str ")")
	      | loop ff (Group(_,x)) = (Str "(")*(loop ff x)*(Str ")")
	      | loop ff  Fake = (Str "<Fake>")
	in loop (fn x=>x) tree end

    (* This will eventually be a pretty printer.
       The `isDecl' is a boolean flag for printing the declaration of an
       object rather than just its name *)
    fun pt2strCommon isDecl tree =
	let val op * = fn (x,y) => Conc(x,y)
	    (* Different info to carry around for pretty printing *)
	    type Cxt = { (* Current substitution from `ObjectInst' *)
			 subst: ParseTree -> ParseTree,
			 (* whether it's a declaration or a use *)
			 isDecl: bool,
			 (* Whether it is a type or not (for type constructors) *)
			 isType: bool,
			 offset: int }
	    val cxt = {subst=(fn x=>x),
		       isDecl=isDecl,
		       isType=false,
		       offset=0}
	    fun upd_subst{subst=ff,
			  isDecl=d,
			  isType=t,
			  offset=n} newff =
		  {subst=newff,
		   isDecl=d,
		   isType=t,
		   offset=n}
	    fun upd_isDecl{subst=ff,
			   isDecl=d,
			   isType=t,
			   offset=n} newd =
		  {subst=ff,
		   isDecl=newd,
		   isType=t,
		   offset=n}
	    fun upd_isType{subst=ff,
			   isDecl=d,
			   isType=t,
			   offset=n} newt =
		  {subst=ff,
		   isDecl=d,
		   isType=newt,
		   offset=n}
	    fun upd_offset{subst=ff,
			   isDecl=d,
			   isType=t,
			   offset=n} offset =
		  {subst=ff,
		   isDecl=d,
		   isType=t,
		   offset=offset}
	    fun add_offset{subst=ff,
			   isDecl=d,
			   isType=t,
			   offset=n} offsetDiff =
		  {subst=ff,
		   isDecl=d,
		   isType=t,
		   offset=n+offsetDiff}
	    fun subst({subst=ff,...}:Cxt)  = ff
	    fun isDecl({isDecl=d,...}:Cxt) = d
	    fun isType({isType=d,...}:Cxt) = d
	    fun offset({offset=n,...}:Cxt) = n
	    fun offsetString({offset=n,...}:Cxt) = 
		 Vector.foldr(op ^) "" (Vector.tabulate(n,(fn _ => " ")))
	    fun offsetStr cxt = Str(offsetString cxt)
	    fun nls cxt = "\n"^(offsetString cxt)
	    fun nl cxt = (Str (nls cxt))
	    fun list2str cxt sep (x::y::tl) = 
		(loop cxt x)*(Str sep)*(list2str cxt sep (y::tl))
	      | list2str cxt _ ([x]) = loop cxt x
	      | list2str _ _ [] = Str ""
	    and option2str _ NONE = Str ""
	      | option2str cxt (SOME tree) = loop cxt tree
	    and optionLst2str _ _ NONE = Str ""
	      | optionLst2str cxt sep (SOME lst) = list2str cxt sep lst
	    and loop (cxt:Cxt) (Program(_,lst)) =
		  (list2str cxt (";"^(nls cxt)) lst)
	      | loop cxt (Include(_,str)) = (Str("include(\""^str^"\")"))
	      | loop cxt (Included(_,str,x)) = 
		(Str("-- include(\""^str^"\")"^(nls cxt)))
		*(loop cxt x)
		*(Str((nls cxt)^"-- end of file "
		      ^str^(nls cxt)))
	      | loop cxt (Quote(_,str)) = (Str ("\""^str^"\""))
	      | loop cxt (SharedRef x) = 
		  (Str "SharedRef(")*(loop cxt (!x))*(Str ")")
	      | loop cxt (Id(_,str)) = Str str
	      | loop cxt (Uid(str)) = Str("<"^str^">")
	      | loop cxt (Tvar(_,str)) = Str("'"^str)
	      | loop cxt (TvarN n) = Str("'X!"^(Int.toString n))
	      | loop cxt (TypeInst(_,parms,tp)) =
		  let val {subst=ff,...} = cxt
		  in (Str "(")*(list2str cxt "," parms)*(Str ") ")
		      *(loop cxt (ff tp))
		  end
	      | loop cxt (Ellipsis _) = Str "..."
	      | loop cxt (TopClosure lst) =
		  let val cxt = upd_isDecl(add_offset cxt 2) true
		  in (Str ((nls cxt)^"-- Top Closure"^(nls cxt)))
		      *(list2str cxt (";"^(nls cxt)) (List.rev lst))
		      *(Str ((nls cxt)^"-- End of Top Closure"^(nls cxt)))
		  end
	      | loop cxt (Module(_,{ name=name,
				   statparams=stat,
				   dynparams=dyn,
				   body=body})) = 
		(Str "module ")*(loop cxt name)
		  *(case stat of
			NONE => Str ""
		      | SOME lst => (Str "[")*(list2str cxt "," lst)*(Str "]"))
		  *(case dyn of
			NONE => Str ""
		      | SOME p => loop cxt p)
		  *(Str " = ")*(loop cxt body)
	      | loop cxt (ModuleClosure{name=name,
					uname=uname,
				       dynName=dn,
				       Sig=ms,
				       def=def,
				       closure=cl,
				       parent=p}) =
		  if isDecl cxt then
		    (Str ((nls cxt)^"module "))
		    *(case name of
			  NONE => (loop cxt uname) (* (Str "<unnamed>") *)
			| SOME n => (loop cxt n))
		    *(Str ": ")*(loop cxt ms)
		    *(Str (" ="))
		    *(case def of
			  NONE =>
			      (nl cxt)
			      *(let val cxt = upd_isDecl(add_offset cxt 2) true
				in (Str ("begin"^(nls cxt)))
				    *(list2str cxt (";"^(nls cxt))
				      (List.rev cl))
				end)*(Str ((nls cxt)^"end"^(nls cxt)))
			| SOME d => (loop (upd_isDecl cxt false) d)*(nl cxt))
		  else (case name of
			  NONE => (Str "<unnamed module>")
			| SOME n => (loop cxt n))
	      | loop cxt (ModuleSig{statparams=stat}) =
		  (Str "ModuleSig{statparams=")
		  *(case stat of
			NONE => Str "[]"
		      | SOME lst => (Str "[")*(list2str cxt "," lst)*(Str "]"))
		  *(Str "}")
	      | loop cxt (ModuleParamClosure{closure=cl, parent=p}) =
		  (Str "ModuleParamClosure")
		  *(nl (add_offset cxt 2))
		  *(Str "{closure = [")
		  *(let val cxt = (add_offset cxt 14)
		    in (list2str cxt (","^(nls cxt)) (List.rev cl))
		    end)
		  *(Str ("],"^(nls (add_offset cxt 3))^"parent = "))
		  (* *(loop cxt p) *) *(Str "<skipping...>")
		  *(Str ("}"^(nls cxt)))
	      | loop cxt (FunClosure{formals=parms,
				     body=body, ...}) =
		  let val cxt = (add_offset cxt 3)
		  in (Str "(fn ")*(loop cxt parms)
		      *(Str " => ")
		      *(loop (add_offset cxt 2) body)*(Str ")")
		  end
	      | loop cxt (RecurFun f) =
		  let val cxt = (add_offset cxt 9)
		  in (Str "RecurFun(")*(loop cxt f)*(Str ")")
		  end
	      | loop cxt (LetClosure{locals=locs,parent=p,body=b}) =
		  (Str ((nls cxt)^"let "))
		   *(let val cxt = upd_isDecl(add_offset cxt 4) true
		     in (list2str cxt (nls cxt) (List.rev locs))
		     end)
		  *(Str ((nls cxt)^"in "))
		   *(loop (upd_isDecl(add_offset cxt 3) false) b)
		  *(Str ((nls cxt)^"end"^(nls cxt)))
	      | loop cxt (LetAsstClosure{locals=locs,parent=p,body=b}) =
		  (Str ((nls cxt)^"let "))
		   *(let val cxt = upd_isDecl(add_offset cxt 4) true
		     in (list2str cxt (nls cxt) (List.rev locs))
		     end)
		  *(Str ((nls cxt)^"in "))
		   *(let val cxt = upd_isDecl(add_offset cxt 3) false
		     in list2str cxt (";"^(nls cxt)) b
		     end)
		  *(Str ((nls cxt)^"end"^(nls cxt)))
	      | loop cxt (TypeClosure{name=name,uname=uname,
				      params=parms,def=def,
				      recursive=r,parent=p}) =
		  (* Reset the substitution - the type closure has no
		     external type vars. *)
		  let val cxt = upd_isType(upd_subst cxt (fn x=>x)) true
		      val {isDecl=d,...} = cxt
		  in 
		      if d then
			  (Str (if r then "datatype " else "type "))
			  *(case parms of
				[] => (Str "")
			      | [a] =>(loop cxt a)
			      | _ => (Str "(")*(list2str cxt "," parms)
				    *(Str ")"))
			  *(Str " ")*(loop cxt name)*(Str " =")
			  *(let val cxt = add_offset (upd_isDecl cxt false) 3
			    in (Str "  ")*(loop cxt def)
			    end)
		      else (loop cxt name)
		  end
	      | loop cxt (ChoiceClosure{pattern=pat,names=names,
					body=b,parent=p,...}) =
		  (loop cxt pat)*(Str " => ")*(loop cxt b)
	      | loop cxt (ChoiceAsstClosure{pattern=pat,
					   names=names,
					   body=b,parent=p,...}) =
		  (loop cxt pat)
		  *(let val cxt = add_offset cxt 3
		    in (Str (" =>"^(nls cxt)))
			*(list2str cxt (nls cxt) b)
		    end)
	      | loop cxt (ChooseClosure{pattern=pat,names=names,
					body=b,parent=p,...}) =
		  (Str "pick ")
		  *(let val cxt = upd_isDecl cxt false
		    in (loop cxt pat)*(Str ": ")*(loop cxt b)
		    end)
	      | loop cxt (ForallClosure{names=names,body=body,parent=p}) =
		  let val cxt = upd_isDecl cxt false
		  in (Str "forall ")
		      *(list2str cxt "," names)*(Str ": ")
		      *(loop cxt body)
		  end
	      | loop cxt (ExistsClosure{names=names,body=body,parent=p}) =
		  let val cxt = upd_isDecl cxt false
		  in (Str "exists ")
		      *(list2str cxt "," names)*(Str ": ")
		      *(loop cxt body)
		  end
	      | loop cxt (QuantifiedVar{name=n,uname=_,Type=tp}) =
		  (Str "(")*(loop cxt n)*(Str ": ")
		  *(loop (upd_isType(upd_isDecl cxt false) true)
		         (subst cxt tp))*(Str ")")
	      | loop cxt (SyncClosure{names=names,body=body,parent=p}) =
		  let val cxt = upd_isDecl cxt false
		  in (Str "sync ")
		      *(list2str cxt "," names)*(Str ": ")
		      *(loop cxt body)
		  end
	      | loop cxt (AsyncClosure{names=names,body=body,parent=p}) =
		  let val cxt = upd_isDecl cxt false
		  in (Str "async ")
		      *(list2str cxt "," names)*(Str ": ")
		      *(loop cxt body)
		  end
	      | loop cxt (Builtin{name=n,Type=tp}) = 
		  if isDecl cxt then (Str "<Builtin(")*(loop cxt n)*(Str ")>")
		  else (loop cxt n)
	      | loop cxt (StateVar{name=name,uname=_,Type=tp,id=id}) =
		  if isDecl cxt then 
		      let val c = upd_isDecl cxt false
		      in  (Str "StateVar ")*(loop c name)
			  *(Str ": ")*(loop (upd_isType c true) (subst c tp))
			  *(Str " (* id=")*(loop c id)*(Str " *)")
		      end
		  else (loop cxt name)
	      | loop cxt (Object{name=name,uname=_,Type=tp,def=def}) =
		  if isDecl cxt then
		      (Str "val ")
		      *(case name of
			    Id _ => (loop cxt name)
			  | _ => (Str "op ")*(loop cxt name))
		      *(Str ": ")
		      *(let val cxt = upd_isDecl(add_offset cxt 5) false
			in (loop (upd_isType(add_offset cxt 3) true) tp)
			    *(Str " =")
			    *(nl cxt)*(loop cxt def)
			end)
		  else (loop cxt name)
	      (* Here we take the new substitution from ObjectInst *) 
	      | loop cxt (ObjectInst{obj=obj,subst=ff1}) =
		  (Str "@(")*
		  (loop (upd_isDecl 
			 (upd_subst cxt ((subst cxt) o ff1)) false) obj)
		  *(Str")")
	      | loop cxt (PatternFormal{name=name,Type=tp,...}) =
		  (Str "<[")*(loop cxt name)*(Str ": ")
		  *(loop (upd_isType cxt true) (subst cxt tp))
		  *(Str "]>")
	      | loop cxt (DynPatternFormal{name=name,Type=tp,...}) =
		  (Str "#<[")*(loop cxt name)*(Str ": ")
		  *(loop (upd_isType cxt true) (subst cxt tp))
		  *(Str "]>")
	      | loop cxt (TypeConstr{name=name,uname=un,Type=tp}) =
		  if isDecl cxt then
		      let val cxt = add_offset cxt 3
		      in (Str "(* ")*(loop cxt name)
			  *(Str ": ")
			  *(loop (upd_isDecl (upd_isType cxt true) false)
			    (subst cxt tp))
			  *(Str " *)")
		      end
		  else (loop cxt name)
		      *(if isType cxt then 
			    (case tp of
				 FunType(_,t,_) => (Str " of ")*(loop cxt t)
			       | _ => Str "")
			else (Str ""))		      
	      | loop cxt (RecordField{name=name,Type=tp}) =
		  (loop cxt (TypedExpr(dp,name,(subst cxt tp))))
	      | loop cxt (StaticFormalConst{name=name,uname=_,Type=tp}) =
		  if isDecl cxt then 
		      (loop (upd_isDecl cxt false)
		            (TypedExpr(dp,name,(subst cxt tp))))
		  else (loop cxt name)
	      | loop cxt (StaticFormalType{name=name,uname=_}) =
		   (Str "type ")*(loop cxt name)
	      | loop cxt (Theorem(_,name,body)) = 
		   let val cxt2 = add_offset cxt 2
		   in (Str "theorem ")*(loop cxt2 name)
		       *(Str (" ="^(nls cxt2)))
		       *(loop (upd_isDecl cxt2 false) body)
		   end
	      | loop cxt (PType(_,tp)) = 
		  (if isDecl cxt then (Str "type ") else (Str ""))
		  *(loop cxt (subst cxt tp))
	      | loop cxt (Symmetric(_,tp)) = 
		  (Str "symmetric ")*(loop cxt (subst cxt tp))
	      | loop cxt (Finite(_,tp)) =
		  (Str "finite ")*(loop cxt (subst cxt tp))
	      | loop cxt (BeginEnd(_,exprs)) = 
		  (Str ((nls cxt)^"begin"))
		  *(let val cxt = add_offset cxt 2
		    in (Str ((nls cxt)^"  "))
			*(list2str cxt (";"^(nls cxt)) exprs)
		    end)
		  *(Str ((nls cxt)^"end -- module"^(nls cxt)))
	      | loop cxt (Sync2(_,e1,e2)) = 
		  (Str "(")*(loop cxt e1)*(Str ")")
		  *(Str " || ")*(Str "(")*(loop cxt e2)*(Str ")")
	      | loop cxt (Async2(_,e1,e2)) = 
		  (Str "(")*(loop cxt e1)*(Str ")")
		  *(Str " | ")*(Str "(")*(loop cxt e2)*(Str ")")
	      | loop cxt (Sync(_,vars,e)) =
		  (Str "sync ")*(list2str cxt ", " vars)*(Str ": ")
		  *(loop cxt e)
	      | loop cxt (Async(_,vars,e)) = 
		  (Str "async ")*(list2str cxt ", " vars)*(Str ": ")
		  *(loop cxt e)
	      | loop cxt (TypeDecl(_,name,parms,tp)) = 
		  (Str "type ")
		  *(case parms of
		       [] => (Str "")
		     | _ => (Str "(")*(list2str cxt "," parms)*(Str ") "))
		  *(loop cxt name)*(Str " = ")
		  *(loop (upd_isType cxt true) (subst cxt tp))
	      | loop cxt (DatatypeDecl(_,name,parms,tp)) = 
		  (Str "datatype ")
		  *(case parms of
		       [] => (Str "")
		     | _ => (Str "(")*(list2str cxt "," parms)*(Str ") "))
		  *(loop cxt name)
		  *(Str " = ")*(loop (upd_isType cxt true) (subst cxt tp))
	      | loop cxt (Var(_,vars)) = 
		  (Str "var ")*(list2str cxt ", " vars)
	      | loop cxt (StateVarDecl(_,vars)) = 
		  (Str "statevar ")*(list2str cxt ", " vars)
	      | loop cxt (Val(_,name,e)) = 
		  (Str "val ")*(loop cxt name)*(Str " = ")*(loop cxt e)
	      | loop cxt (FunCases(_,lst)) = 
		  let val cxt = add_offset cxt 2
		  in (Str "fun ")
		      *(list2str (add_offset cxt 2) ((nls cxt)^"| ") lst)
		  end
	      | loop cxt (FunCase(_,name,patterns,res_tp,e)) = 
		  (loop cxt name)*(Str " ")*(list2str cxt " " patterns)
		  *(case res_tp of
			NONE => Str ""
		      | SOME tp => (Str " : ")
			    *(loop (upd_isType cxt true) (subst cxt tp)))
		  *(Str " = ")
		  *(let val cxt = add_offset cxt 3
		    in (nl cxt)*(loop cxt e)
		    end)
	      | loop cxt (Open(_,e)) = (Str "open ")*(loop cxt e)
	      | loop cxt (Export(_,lst)) = 
		  (Str "export ")
		  *(case lst of
			[] => (Str "all")
		      | _  => list2str cxt ", " lst)
	      | loop cxt (Models(_,e1,e2)) = 
		  (loop cxt e1)*(Str ((nls cxt)^"|="^(nls cxt)))
		  *(loop cxt e2)
	      | loop cxt (Refines(_,e1,e2)) = 
		  (loop cxt e1)*(Str ((nls cxt)^"refines"^(nls cxt)))
		  *(loop cxt e2)
	      | loop cxt (Nop _) = Str "nop"
	      | loop cxt (NormalAsst(_,name,e)) = 
		  (Str "(")*(loop cxt name)*(Str ") := ")
		  *(loop (upd_isDecl cxt false) e)
	      | loop cxt (InitAsst(_,name,e)) = 
		  (Str "init(")*(loop cxt name)*(Str ") := ")
		  *(loop (upd_isDecl cxt false) e)
	      | loop cxt (NextAsst(_,name,e)) = 
		  (Str "next(")*(loop cxt name)*(Str ") := ")
		  *(loop (upd_isDecl cxt false) e)
	      | loop cxt (Next e) = (Str "next(")*(loop cxt e)*(Str ")")
	      | loop cxt (IfAsst(_,cond_lst,else_lst)) = 
		  let val cxt2 = add_offset cxt 2
		  in (Str "if ")
		      *(list2str cxt2 ((nls cxt)^"elsif ") cond_lst)
		      *(Str ((nls cxt)^"else"^(nls cxt2)))
		      *(list2str cxt2 (nls cxt2) else_lst)
		      *(Str ((nls cxt)^"endif"))
		  end
	      | loop cxt (CondAsst(_,e,assts)) = 
		  (loop cxt e)*(Str (" then"^(nls cxt)))
		  *(list2str cxt (nls cxt) assts)
	      | loop cxt (LetAsst(_,locals,assts)) = 
		  (Str "let ")
		  *(let val cxt = upd_isDecl(add_offset cxt 4) true
		    in (list2str cxt (nls cxt) locals)
		    end)
		  *(Str ((nls cxt)^"in"^(nls cxt)^"  "))
		  *(let val cxt = add_offset cxt 3
		    in (list2str cxt (";"^(nls cxt)) assts)
		    end)
		  *(Str ((nls cxt)^"end "))
	      | loop cxt (CaseAsst(_,e,cases)) = 
		  let val cxt2 = add_offset cxt 2
		      val cxt4 = add_offset cxt 4
		  in (Str ((nls cxt)^"case "))*(loop cxt4 e)*(Str (" of"^(nls cxt4)))
		      *(list2str cxt4 ((nls cxt2)^"| ") cases)
		      *(Str ((nls cxt)^"endcase"))
		  end
	      | loop cxt (ChoiceAsst(_,e,assts)) =
		  let val cxt2 = add_offset cxt 2
		      val cxt6 = add_offset cxt 6
		  in (loop cxt2 e)
		      *(Str (" =>"^(nls cxt6)))
		      *(list2str cxt6 (";"^(nls cxt6)) assts)
		  end
	      | loop cxt (ChooseAsst(_,lst)) = 
		  let val cxt2 = add_offset cxt 2
		      val cxt4 = add_offset cxt 4
		  in (Str ("choose"^(nls cxt4)))
		      *(list2str cxt ((nls cxt2)^"| ") lst)
		      *(Str ((nls cxt)^"endchoose"))
		  end
	      | loop cxt (IfExpr(_,conds,else_exp)) = 
		  let val cxt2 = add_offset cxt 2
		  in (Str "if ")*(list2str cxt2 ((nls cxt)^"elsif ") conds)
		      *(Str ((nls cxt)^"else "))*(loop cxt2 else_exp)
		      *(Str ((nls cxt)^"endif "))
		  end
	      | loop cxt (CondExpr(_,e1,e2)) = 
		  (loop (add_offset cxt 3) e1)
		  *(Str (" then"^(nls cxt)))*(loop cxt e2)
	      | loop cxt (LetExpr(_,locals,e)) =
		  (Str "let ")
		  *(let val cxt = upd_isDecl(add_offset cxt 4) true
		    in (list2str cxt (nls cxt) locals)
		    end)
		  *(Str ((nls cxt)^"in "))
		  *(loop (upd_isDecl(add_offset cxt 3) false) e)
		  *(Str ((nls cxt)^"end "))
	      | loop cxt (CaseExpr(_,e,cases)) = 
		  let val cxt2 = add_offset cxt 2
		      val cxt4 = add_offset cxt 4
		  in (Str ((nls cxt)^"case "))*(loop cxt4 e)
		      *(Str (" of"^(nls cxt4)))
		      *(list2str cxt4 ((nls cxt2)^"| ") cases)
		  end
		  *(Str ((nls cxt)^"endcase "))
	      | loop cxt (ChoiceExpr(_,p,e)) = 
		  let val cxt2 = add_offset cxt 2
		      val cxt6 = add_offset cxt 6
		  in (loop cxt2 p)*(Str (" =>"^(nls cxt6)))*(loop cxt6 e)
		  end
	      | loop cxt (ChooseExpr(_,p,e)) = 
		  (Str "pick ")*(loop cxt p)*(Str ": ")*(loop cxt e)
	      | loop cxt (WithExpr(_,e,lst)) = 
		  (loop cxt e)*(Str " with [ ")*(list2str cxt ", " lst)*(Str " ]")
	      | loop cxt (WithAsst(_,name,e)) = 
		  (loop cxt name)*(Str " = ")*(loop cxt e)
	      | loop cxt (RecordExpr(_,lst)) = 
		  (Str "{ ")*(list2str cxt ", " lst)*(Str " }")
	      | loop cxt (TupleExpr(_,lst)) = 
		  (Str "( ")*(list2str cxt ", " lst)*(Str " )")
	      | loop cxt (NondetExpr(_,lst)) = 
		  (Str "(")*(list2str cxt ") | (" lst)*(Str ")")
	      | loop cxt (RecordAsst(_,name,e)) = 
		  (loop cxt name)*(Str " = ")*(loop cxt e)
	      | loop cxt (ModuleInst(_,name,stats)) = 
		  (loop cxt name)*(Str "[")*(list2str cxt ", " stats)*(Str "]")
	      | loop cxt (Dot(_,e1,e2)) = 
		  (loop cxt e1)*(Str ".")*(loop cxt e2)
	      | loop cxt (Type(_,e)) = 
		  (Str "type ")*(loop cxt e)
	      | loop cxt (Datatype(_,e)) = 
		  (Str "datatype ")*(loop cxt e)
	      | loop cxt (And _) = (Str "and")
	      | loop cxt (Or _) = (Str "or")
	      | loop cxt (Implies _) = (Str "implies")
	      | loop cxt (Arrow _) = (Str "->")
	      | loop cxt (Darrow _) = (Str "=>")
	      | loop cxt (Iff _) = (Str "<=>")
	      | loop cxt (Eq _) = (Str "=")
	      | loop cxt (NotEq _) = (Str "!=")
	      | loop cxt (Not _) = (Str "!")
	      | loop cxt (Forall(_,vars,e)) = 
		  (Str "forall ")*(list2str cxt ", " vars)*(Str ": ")*(loop cxt e)
	      | loop cxt (Exists(_,vars,e)) = 
		  (Str "exists ")*(list2str cxt ", " vars)*(Str ": ")*(loop cxt e)
	      | loop cxt (Ag _) = (Str "AG")
	      | loop cxt (Af _) = (Str "AF")
	      | loop cxt (Eg _) = (Str "EG")
	      | loop cxt (Ef _) = (Str "EF")
	      | loop cxt (Ax _) = (Str "AX")
	      | loop cxt (Ex _) = (Str "EX")
	      | loop cxt (Au _) = (Str "AU")
	      | loop cxt (Eu _) = (Str "EU")
	      | loop cxt (Ar _) = (Str "AR")
	      | loop cxt (Er _) = (Str "ER")
	      | loop cxt (Lt _) = (Str "<")
	      | loop cxt (Gt _) = (Str ">")
	      | loop cxt (Le _) = (Str "<=")
	      | loop cxt (Ge _) = (Str ">=")
	      | loop cxt (Plus _) = (Str "+")
	      | loop cxt (Minus _) = (Str "-")
	      | loop cxt (UMinus _) = (Str "~")
	      | loop cxt (Times _) = (Str "*")
	      | loop cxt (Div _) = (Str "/")
	      | loop cxt (Mod _) = (Str "mod")
	      | loop cxt (Number(_,n)) = (Str (Int.toString n))
	      | loop cxt (BoolType _) = (Str "bool")
	      | loop cxt (BoolType2) = (Str "bool2")
	      | loop cxt (NatType _) = (Str "nat")
	      | loop cxt (IntType _) = (Str "int")
	      | loop cxt (RangeType(_,x,y)) =
		  (Str "[")*(loop cxt x)*(Str "..")*(loop cxt y)*(Str "]")
	      | loop cxt (FunType(_,tp1,tp2)) =
		  (Str "(")*(loop cxt (subst cxt tp1))*(Str ")")
		  *(Str " -> ")*(Str "(")*(loop cxt (subst cxt tp2))*(Str ")")
	      | loop cxt (TupleType(_,lst)) = 
		  (Str "(")*(list2str cxt ") * (" lst)*(Str ")")
	      | loop cxt (ArrayType(_,tp1,tp2)) = 
		  (Str "array ")*(loop cxt (subst cxt tp1))*(Str " of ")
		  *(loop cxt (subst cxt tp2))
	      | loop cxt (RecordType(_,lst)) = 
		  (Str "{ ")*(list2str cxt ", " lst)*(Str " }")
	      | loop cxt (EnumType(_,lst)) = 
		  (Str "(")*(list2str cxt " | " lst)*(Str ")")
	      | loop cxt (AbstractType(lst)) = 
		  (Str "AbstractType[")*(list2str cxt ", " lst)*(Str "]")
	      | loop cxt (Of(_,name,tp)) = 
		  (loop cxt name)*(Str " of ")*(loop cxt (subst cxt tp))
	      | loop cxt (TypedVars(_,vars,tp)) = 
		  (Str "(")*(list2str cxt ", " vars)
		  *(case tp of
			NONE => (Str "")
		      | SOME tp => (Str ": ")
			    *(loop (upd_isType cxt true) (subst cxt tp)))
		  *(Str ")")
	      | loop cxt (False _) = (Str "false")
	      | loop cxt (True _) = (Str "true")
	      | loop cxt (False2) = (Str "false2")
	      | loop cxt (True2) = (Str "true2")
	      | loop cxt (Min _) = (Str "min")
	      | loop cxt (Max _) = (Str "max")
	      | loop cxt (Undefined _) = (Str "undefined")
	      | loop cxt (Anyvalue _) = (Str "anyvalue")
	      | loop cxt OtherValue = (Str "OtherValue")
	      | loop cxt (Self _) = (Str "self")
	      | loop cxt (Appl(_,Au _,TupleExpr(_,[e1,e2]))) =
		  (Str "Always[")*(loop cxt e1)*(Str " Until ")*(loop cxt e2)*(Str "]")
	      | loop cxt (Appl(_,Eu _,TupleExpr(_,[e1,e2]))) =
		  (Str "Sometimes[")*(loop cxt e1)*(Str " Until ")
		  *(loop cxt e2)*(Str "]")
	      | loop cxt (Appl(_,Ar _,TupleExpr(_,[e1,e2]))) =
		  (Str "Always[")*(loop cxt e1)*(Str " Releases ")
		  *(loop cxt e2)*(Str "]")
	      | loop cxt (Appl(_,Er _,TupleExpr(_,[e1,e2]))) =
		  (Str "Sometimes[")*(loop cxt e1)*(Str " Releases ")
		  *(loop cxt e2)*(Str "]")
	      | loop cxt (Appl(_,e1,e2)) = 
		  let val cxt = upd_isDecl cxt false
		  in if isInfix(e1) then
		       (case e2 of
			    TupleExpr(_,[x,y]) =>
			      (Str "(")*(loop cxt x)*(Str " ")
			      *(loop cxt e1)*(Str " ")*(loop cxt y)*(Str ")")
			  | x => (Str "(op ")*(loop cxt e1)*(Str " ")
			      *(loop cxt e2)*(Str ")"))
		     else if isUnary(e1) then
		      (loop cxt e1)*(Str "(")*(loop cxt e2)*(Str ")")
		     else
		      (Str "(")*(loop cxt e1)*(Str " ")*(loop cxt e2)*(Str ")")
		  end
	      | loop cxt (TypedExpr(_,e,tp)) = 
		  (Str "((")*(loop cxt e)*(Str "): ")
		  *(loop (upd_isType cxt true) (subst cxt tp))*(Str ")")
	      | loop cxt (Fn(_,lst)) =
		  let val cxt2 = add_offset cxt 2
		      val cxt4 = add_offset cxt 4
		  in (Str "fn ")*(list2str cxt4 ((nls cxt2)^"| ") lst)
		  end
	      | loop cxt (ChoicePattern(_,lst)) = (list2str cxt " | " lst)
	      | loop cxt (ApplPattern(_,p1,p2)) =
		  (Str "(")*(loop cxt p1)*(Str " ")*(loop cxt p2)*(Str ")")
	      | loop cxt (AsPattern(_,p1,p2)) =
		  (Str "(")*(loop cxt p1)*(Str " as ")*(loop cxt p2)*(Str ")")
	      | loop cxt (TypedPattern(_,p,tp)) = 
		  (Str "(")*(loop cxt p)*(Str ": ")*(loop cxt (subst cxt tp))*(Str ")")
	      | loop cxt (TuplePattern(_,lst)) = 
		  (Str "(")*(list2str cxt ", " lst)*(Str ")")
	      | loop cxt (RecordPattern(_,lst)) = 
		  (Str "{")*(list2str cxt ", " lst)*(Str "}")
	      | loop cxt (Underscore _) = (Str "_")
	      | loop cxt (ExtractAppl(c,e)) =
		  (Str "ExtractAppl(")*(list2str cxt ", " [c,e])
		  *(Str ")")
	      | loop cxt (ExtractTuple(n,e)) =
		  (Str("ExtractTuple("^(Int.toString n)^","))
		  *(loop cxt e)*(Str ")")
	      | loop cxt (ExtractRecord(name,e)) =
		  (Str "ExtractRecord(")*(loop cxt name)*(Str ",")
		  *(loop cxt e)*(Str ")")
	      | loop cxt (ExtractIndex e) =
		  (Str "ExtractIndex(")*(loop cxt e)*(Str ")")
	      | loop cxt (Group(_,x)) = (Str "(")*(loop cxt x)*(Str ")")
	      | loop cxt  Fake = (Str "<Fake>")
(*  	      | loop cxt  x = raise SympBug("Can't print a ParseTree element.") *)
	in loop cxt tree
	end

    fun pt2str tree = pt2strCommon false tree

    fun pt2string tree = Str2string(pt2str tree)

    fun pt2stringDebug tree = Str2string(pt2strDebug tree)

    fun pt2shortStr length tree =
	let val str = pt2string tree
	in if (String.size str) > length then
	     (String.substring(str,0,length-3))^"..."
	   else str
	end

    (* Create a string from the list of objects separated with sep *)
    fun ptlist2str sep [] = ""
      | ptlist2str sep [hd] = pt2string hd
      | ptlist2str sep (hd::tl) = (pt2string hd)^sep^(ptlist2str sep tl)

    fun ptlist2strDebug sep [] = ""
      | ptlist2strDebug sep [hd] = pt2stringDebug hd
      | ptlist2strDebug sep (hd::tl) =
	  (pt2stringDebug hd)^sep^(ptlist2strDebug sep tl)

    fun pos(Program(p,_)) = p
      | pos(Include(p,_)) = p
      | pos(Included(p,_,_)) = p
      | pos(Quote(p,_)) = p
      | pos(Id(p,_)) = p
      | pos(Tvar(p,_)) = p
      | pos(Ellipsis p) = p
      | pos(Module(p,_)) = p
      | pos(Theorem(p,_,_)) = p
      | pos(PType(p,_)) = p
      | pos(Symmetric(p,_)) = p
      | pos(Finite(p,_)) = p
      | pos(BeginEnd(p,_)) = p
      | pos(Sync2(p,_,_)) = p
      | pos(Async2(p,_,_)) = p
      | pos(Sync(p,_,_)) = p
      | pos(Async(p,_,_)) = p
      | pos(TypeDecl(p,_,_,_)) = p
      | pos(DatatypeDecl(p,_,_,_)) = p
      | pos(TypeInst(p,_,_)) = p
      | pos(Var(p,_)) = p
      | pos(StateVarDecl(p,_)) = p
      | pos(Val(p,_,_)) = p
      | pos(FunCases(p,_)) = p
      | pos(FunCase(p,_,_,_,_)) = p
      | pos(Open(p,_)) = p
      | pos(Export(p,_)) = p
      | pos(Models(p,_,_)) = p
      | pos(Refines(p,_,_)) = p
      | pos(Nop p) = p
      | pos(NormalAsst(p,_,_)) = p
      | pos(InitAsst(p,_,_)) = p
      | pos(NextAsst(p,_,_)) = p
      | pos(IfAsst(p,_,_)) = p
      | pos(CondAsst(p,_,_)) = p
      | pos(LetAsst(p,_,_)) = p
      | pos(CaseAsst(p,_,_)) = p
      | pos(ChoiceAsst(p,_,_)) = p
      | pos(ChooseAsst(p,_)) = p
      | pos(IfExpr(p,_,_)) = p
      | pos(CondExpr(p,_,_)) = p
      | pos(LetExpr(p,_,_)) = p
      | pos(CaseExpr(p,_,_)) = p
      | pos(ChoiceExpr(p,_,_)) = p
      | pos(ChooseExpr(p,_,_)) = p
      | pos(WithExpr(p,_,_)) = p
      | pos(WithAsst(p,_,_)) = p
      | pos(RecordExpr(p,_)) = p
      | pos(TupleExpr(p,_)) = p
      | pos(NondetExpr(p,_)) = p
      | pos(RecordAsst(p,_,_)) = p
      | pos(ModuleInst(p,_,_)) = p
      | pos(Dot(p,_,_)) = p
      | pos(Type(p,_)) = p
      | pos(Datatype(p,_)) = p
      | pos(And(p)) = p
      | pos(Or(p)) = p
      | pos(Implies(p)) = p
      | pos(Iff(p)) = p
      | pos(Eq(p)) = p
      | pos(NotEq(p)) = p
      | pos(Not(p)) = p
      | pos(Forall(p,_,_)) = p
      | pos(Exists(p,_,_)) = p
      | pos(Ag(p)) = p
      | pos(Af(p)) = p
      | pos(Eg(p)) = p
      | pos(Ef(p)) = p
      | pos(Ax(p)) = p
      | pos(Ex(p)) = p
      | pos(Au(p)) = p
      | pos(Eu(p)) = p
      | pos(Ar(p)) = p
      | pos(Er(p)) = p
      | pos(Lt(p)) = p
      | pos(Gt(p)) = p
      | pos(Le(p)) = p
      | pos(Ge(p)) = p
      | pos(Plus(p)) = p
      | pos(Minus(p)) = p
      | pos(UMinus(p)) = p
      | pos(Times(p)) = p
      | pos(Div(p)) = p
      | pos(Mod(p)) = p
      | pos(Number(p,_)) = p
      | pos(BoolType p) = p
      | pos(NatType p) = p
      | pos(IntType p) = p
      | pos(RangeType(p,_,_)) = p
      | pos(FunType(p,_,_)) = p
      | pos(TupleType(p,_)) = p
      | pos(ArrayType(p,_,_)) = p
      | pos(RecordType(p,_)) = p
      | pos(EnumType(p,_)) = p
      | pos(Of(p,_,_)) = p
      | pos(TypedVars(p,_,_)) = p
      | pos(False p) = p
      | pos(True p) = p
      | pos(Min p) = p
      | pos(Max p) = p
      | pos(Undefined p) = p
      | pos(Anyvalue p) = p
      | pos(Self p) = p
      | pos(Appl(p,_,_)) = p
      | pos(TypedExpr(p,_,_)) = p
      | pos(Fn(p,_)) = p
      | pos(ChoicePattern(p,_)) = p
      | pos(ApplPattern(p,_,_)) = p
      | pos(AsPattern(p,_,_)) = p
      | pos(TypedPattern(p,_,_)) = p
      | pos(TuplePattern(p,_)) = p
      | pos(RecordPattern(p,_)) = p
      | pos(Underscore p) = p
      | pos(Group(p,_)) = p
      | pos x = dp (* Return an "unknown" position *)

(*  raise SympParseInternalError("pos: Found illegal value: "^(pt2string x)) *)

    fun pos2string (l,c) = (Int.toString l)^"."^(Int.toString c)

    (* Equality of two parse trees upto the position field.
     * If noParent=true, then don't recurse into parent closures either,
     * consider them irrelevant for equality. *)
    fun ptEqCommon noParent (t1,t2) =
	let fun looplist(hd1::tl1,hd2::tl2) = 
	           (loop(hd1,hd2)) andalso (looplist(tl1,tl2))
	      | looplist([],[]) = true
	      | looplist _ = false
	    and loop(Program(_,lst1),Program(_,lst2)) = looplist(lst1,lst2)
	      | loop(Include(_,s1),Include(_,s2)) = (s1 = s2)
	      | loop(Included(_,s1,t1),Included(_,s2,t2)) =
		   (s1 = s2) andalso (loop(t1,t2))
	      | loop(Quote(_,s1),Quote(_,s2)) = (s1 = s2)
	      | loop(SharedRef x,SharedRef y) = (x=y)
	      | loop(Id(_,s1),Id(_,s2)) = (s1 = s2)
	      | loop(Uid s1,Uid s2) = (s1 = s2)
	      | loop(Tvar(_,s1),Tvar(_,s2)) = (s1 = s2)
	      | loop(TvarN n1,TvarN n2) = (n1 = n2)
	      | loop(TypeInst(_,lst1,x1), TypeInst(_,lst2,x2)) =
		   looplist(lst1,lst2) andalso loop(x1,x2)
	      | loop(Ellipsis _,Ellipsis _) = true
	      | loop(TopClosure lst1, TopClosure lst2) = looplist(lst1,lst2)
	      | loop(Module(_,{name=name1,statparams=s1,dynparams=d1,
			       body=b1}),
		     Module(_,{name=name2,statparams=s2,dynparams=d2,
			       body=b2})) =
		    (loop(name1,name2))
		    andalso (case (s1,s2) of
				 (NONE,NONE) => true
			       | (SOME(x1),SOME(x2)) => looplist(x1,x2)
			       | _ => false)
		    andalso (case (d1,d2) of
				 (NONE,NONE) => true
			       | (SOME(x1),SOME(x2)) => loop(x1,x2)
			       | _ => false)
		    andalso (loop(b1,b2))
	      | loop(ModuleClosure{name=n1,uname=un1,dynName=dn1,Sig=s1,def=d1,
				   closure=c1,parent=p1},
		     ModuleClosure{name=n2,uname=un2,dynName=dn2,Sig=s2,def=d2,
				   closure=c2,parent=p2}) =
		    (case (n1,n2) of
			 (NONE,NONE) => true
		       | (SOME n1,SOME n2) => (loop(n1,n2))
		       | _ => false)
		    andalso (loop(un1,un2))
		    andalso (case (dn1,dn2) of
				 (NONE,NONE) => true
			       | (SOME d1,SOME d2) => (loop(d1,d2))
			       | _ => false)
		    andalso (loop(s1,s2))
		    andalso (case (d1,d2) of
				 (NONE,NONE) => true
			       | (SOME d1,SOME d2) => (loop(d1,d2))
			       | _ => false)
		    andalso (looplist(c1,c2))
		    andalso (noParent orelse loop(p1,p2))
	      | loop(ModuleSig{statparams=s1},ModuleSig{statparams=s2}) =
		    (case (s1,s2) of
			 (NONE,NONE) => true
		       | (SOME(x1),SOME(x2)) => looplist(x1,x2)
		       | _ => false)
	      | loop(ModuleParamClosure{closure=lst1,parent=p1},
		     ModuleParamClosure{closure=lst2,parent=p2}) =
		    looplist(lst1,lst2) andalso (noParent orelse loop(p1,p2))
	      | loop(FunClosure{name=n1,formals=f1,parent=p1,body=b1},
		     FunClosure{name=n2,formals=f2,parent=p2,body=b2}) =
		    (case (n1,n2) of
			 (NONE,NONE) => true
		       | (SOME(x1),SOME(x2)) => (loop(x1,x2))
		       | _ => false)
		    andalso (loop(f1,f2))
		    andalso (loop(b1,b2))
		    andalso (noParent orelse (loop(p1,p2)))
	      | loop(RecurFun x1, RecurFun x2) = loop(x1,x2)
	      | loop(LetClosure{locals=lst1,parent=p1,body=b1},
		     LetClosure{locals=lst2,parent=p2,body=b2}) =
		    looplist(lst1,lst2) andalso loop(b1,b2) 
		    andalso (noParent orelse loop(p1,p2))
	      | loop(LetAsstClosure{locals=lst1,parent=p1,body=b1},
		     LetAsstClosure{locals=lst2,parent=p2,body=b2}) =
		    looplist(lst1,lst2) andalso looplist(b1,b2)
		    andalso (noParent orelse loop(p1,p2))
	      (* This one might be depricated soon *)
	      | loop(TypeClosure{name=n1,uname=un1,params=lst1,def=d1,
				 recursive=r1,parent=p1},
		     TypeClosure{name=n2,uname=un2,params=lst2,def=d2,
				 recursive=r2,parent=p2}) =
		    loop(n1,n2) andalso loop(un1,un2) andalso r1 = r2
		    andalso looplist(lst1,lst2) andalso loop(d1,d2)
		    andalso (noParent orelse loop(p1,p2))
	      | loop(ChoiceClosure{pattern=pat1,uname=u1,
				   names=n1,body=b1,parent=p1},
		     ChoiceClosure{pattern=pat2,uname=u2,
				   names=n2,body=b2,parent=p2}) =
		    loop(pat1,pat2) andalso looplist(n1,n2)
		    andalso loop(u1,u2) andalso loop(b1,b2)
		    andalso (noParent orelse loop(p1,p2))
	      | loop(ChoiceAsstClosure{pattern=pat1,uname=u1,names=n1,
				       body=b1,parent=p1},
		     ChoiceAsstClosure{pattern=pat2,uname=u2,names=n2,
				       body=b2,parent=p2}) =
		    loop(pat1,pat2) andalso looplist(n1,n2)
		    andalso loop(u1,u2) andalso looplist(b1,b2)
		    andalso (noParent orelse loop(p1,p2))
	      | loop(ChooseClosure{pattern=pat1,Type=tp1,names=n1,body=b1,parent=p1},
		     ChooseClosure{pattern=pat2,Type=tp2,names=n2,body=b2,parent=p2}) =
		    loop(pat1,pat2) andalso looplist(n1,n2)
		    andalso loop(tp1,tp2) andalso loop(b1,b2)
		    andalso (noParent orelse loop(p1,p2))
	      | loop(ForallClosure{names=n1,body=b1,parent=p1},
		     ForallClosure{names=n2,body=b2,parent=p2}) =
		    looplist(n1,n2) andalso loop(b1,b2)
		    andalso (noParent orelse loop(p1,p2))
	      | loop(ExistsClosure{names=n1,body=b1,parent=p1},
		     ExistsClosure{names=n2,body=b2,parent=p2}) =
		    looplist(n1,n2) andalso loop(b1,b2)
		    andalso (noParent orelse loop(p1,p2))
	      | loop(QuantifiedVar{name=n1,uname=un1,Type=tp1},
		     QuantifiedVar{name=n2,uname=un2,Type=tp2}) =
		    loop(n1,n2) andalso loop(un1,un2) andalso loop(tp1,tp2)
	      | loop(SyncClosure{names=n1,body=b1,parent=p1},
		     SyncClosure{names=n2,body=b2,parent=p2}) =
		    looplist(n1,n2) andalso loop(b1,b2)
		    andalso (noParent orelse loop(p1,p2))
	      | loop(AsyncClosure{names=n1,body=b1,parent=p1},
		     AsyncClosure{names=n2,body=b2,parent=p2}) =
		    looplist(n1,n2) andalso loop(b1,b2)
		    andalso (noParent orelse loop(p1,p2))
	      | loop(Builtin{name=n1,Type=tp1},Builtin{name=n2,Type=tp2}) =
		    loop(n1,n2) andalso loop(tp1,tp2)
	      | loop(StateVar{name=n1,uname=un1,Type=tp1,id=id1},
		     StateVar{name=n2,uname=un2,Type=tp2,id=id2}) =
		    (loop(n1,n2)) andalso loop(un1,un2) andalso (loop(tp1,tp2))
		    andalso (loop(id1,id2))
	      | loop(Object{name=n1,uname=un1,Type=t1,def=d1},
		     Object{name=n2,uname=un2,Type=t2,def=d2}) =
		    (loop(n1,n2)) andalso(loop(un1,un2)) andalso (loop(t1,t2))
		    andalso (loop(d1,d2))
	      | loop(ObjectInst{obj=obj1,...},
		     ObjectInst{obj=obj2,...}) = loop(obj1,obj2)
	      | loop(PatternFormal{name=n1,uname=un1,Type=t1,...},
		     PatternFormal{name=n2,uname=un2,Type=t2,...}) =
		    (loop(n1,n2)) andalso (loop(un1,un2)) andalso (loop(t1,t2))
	      | loop(DynPatternFormal{name=n1,uname=un1,Type=t1,...},
		     DynPatternFormal{name=n2,uname=un2,Type=t2,...}) =
		    (loop(n1,n2)) andalso (loop(un1,un2)) andalso (loop(t1,t2))
	      | loop(TypeConstr{name=n1,uname=un1,Type=t1},
		     TypeConstr{name=n2,uname=un2,Type=t2}) =
		    (loop(n1,n2)) andalso (loop(un1,un2)) andalso (loop(t1,t2))
	      | loop(RecordField{name=n1,Type=f1},
		     RecordField{name=n2,Type=f2}) =
		    (loop(n1,n2))  andalso (loop(f1,f2))
	      | loop(StaticFormalConst{name=n1,uname=un1,Type=d1},
		     StaticFormalConst{name=n2,uname=un2,Type=d2}) =
		    (loop(n1,n2)) andalso (loop(un1,un2)) andalso (loop(d1,d2))
	      | loop(StaticFormalType{name=n1,uname=un1},
		     StaticFormalType{name=n2,uname=un2}) =
		    (loop(n1,n2)) andalso (loop(un1,un2))
	      | loop(Theorem(_,x1,y1),Theorem(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(PType(_,x1),PType(_,x2)) = loop(x1,x2)
	      | loop(Symmetric(_,x1),Symmetric(_,x2)) = loop(x1,x2)
	      | loop(Finite(_,x1),Finite(_,x2)) = loop(x1,x2)
	      | loop(BeginEnd(_,lst1),BeginEnd(_,lst2)) = looplist(lst1,lst2)
	      | loop(Sync2(_,x1,y1),Sync2(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(Async2(_,x1,y1),Async2(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(Sync(_,x1,y1),Sync(_,x2,y2)) =
		    (looplist(x1,x2)) andalso (loop(y1,y2))
	      | loop(Async(_,x1,y1),Async(_,x2,y2)) =
		    (looplist(x1,x2)) andalso (loop(y1,y2))
	      | loop(TypeDecl(_,x1,lst1,y1),TypeDecl(_,x2,lst2,y2)) =
		    (loop(x1,x2)) andalso (looplist(lst1,lst2))
		    andalso (loop(y1,y2))
	      | loop(DatatypeDecl(_,x1,lst1,y1),DatatypeDecl(_,x2,lst2,y2)) =
		    (loop(x1,x2)) andalso (looplist(lst1,lst2))
		    andalso (loop(y1,y2))
	      | loop(Var(_,x1),Var(_,x2)) = looplist(x1,x2)
	      | loop(StateVarDecl(_,x1),StateVarDecl(_,x2)) = looplist(x1,x2)
	      | loop(Val(_,x1,y1),Val(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(FunCases(_,x1),FunCases(_,x2)) = looplist(x1,x2)
	      | loop(FunCase(_,x1,lst1,o1,y1),FunCase(_,x2,lst2,o2,y2)) =
		    (loop(x1,x2)) andalso (looplist(lst1,lst2))
		    andalso (case (o1,o2) of
				 (NONE,NONE) => true
			       | (SOME(z1),SOME(z2)) => loop(z1,z2)
			       | _ => false)
		    andalso (loop(y1,y2))
	      | loop(Open(_,x1),Open(_,x2)) = loop(x1,x2)
	      | loop(Export(_,lst1),Export(_,lst2)) = looplist(lst1,lst2)
	      | loop(Models(_,x1,y1),Models(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(Refines(_,x1,y1),Refines(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(Nop _,Nop _) = true
	      | loop(NormalAsst(_,x1,y1),NormalAsst(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(InitAsst(_,x1,y1),InitAsst(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(NextAsst(_,x1,y1),NextAsst(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(Next x1, Next x2) = loop(x1,x2)
	      | loop(IfAsst(_,x1,y1),IfAsst(_,x2,y2)) =
		    (looplist(x1,x2)) andalso (looplist(y1,y2))
	      | loop(CondAsst(_,x1,y1),CondAsst(_,x2,y2)) =
		    (loop(x1,x2)) andalso (looplist(y1,y2))
	      | loop(LetAsst(_,x1,y1),LetAsst(_,x2,y2)) =
		    (looplist(x1,x2)) andalso (looplist(y1,y2))
	      | loop(CaseAsst(_,x1,y1),CaseAsst(_,x2,y2)) =
		    (loop(x1,x2)) andalso (looplist(y1,y2))
	      | loop(ChoiceAsst(_,x1,y1),ChoiceAsst(_,x2,y2)) =
		    (loop(x1,x2)) andalso (looplist(y1,y2))
	      | loop(ChooseAsst(_,lst1),ChooseAsst(_,lst2)) = 
		    looplist(lst1,lst2)
	      | loop(IfExpr(_,lst1,x1),IfExpr(_,lst2,x2)) =
		    (looplist(lst1,lst2)) andalso (loop(x1,x2))
	      | loop(CondExpr(_,x1,y1),CondExpr(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(LetExpr(_,lst1,x1),LetExpr(_,lst2,x2)) =
		    (looplist(lst1,lst2)) andalso (loop(x1,x2))
	      | loop(CaseExpr(_,x1,lst1),CaseExpr(_,x2,lst2)) =
		    (looplist(lst1,lst2)) andalso (loop(x1,x2))
	      | loop(ChoiceExpr(_,x1,y1),ChoiceExpr(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(ChooseExpr(_,x1,y1),ChooseExpr(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(WithExpr(_,x1,lst1),WithExpr(_,x2,lst2)) =
		    (loop(x1,x2)) andalso (looplist(lst1,lst2))
	      | loop(WithAsst(_,x1,y1),WithAsst(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(RecordExpr(_,lst1),RecordExpr(_,lst2)) =
		    (looplist(lst1,lst2))
	      | loop(TupleExpr(_,lst1),TupleExpr(_,lst2)) =
		    (looplist(lst1,lst2))
	      | loop(NondetExpr(_,lst1),NondetExpr(_,lst2)) =
		    (looplist(lst1,lst2))
	      | loop(RecordAsst(_,x1,y1),RecordAsst(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(ModuleInst(_,x1,lst1),ModuleInst(_,x2,lst2)) =
		    (loop(x1,x2)) andalso (looplist(lst1,lst2))
	      | loop(Dot(_,x1,y1),Dot(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(Type(_,x1),Type(_,x2)) = (loop(x1,x2))
	      | loop(Datatype(_,x1),Datatype(_,x2)) = (loop(x1,x2))
	      | loop(And _,And _) = true
	      | loop(Or _,Or _) = true
	      | loop(Implies _,Implies _) = true
	      | loop(Arrow _, Arrow _) = true
	      | loop(Darrow _, Darrow _) = true
	      | loop(Iff _,Iff _) = true
	      | loop(Eq _,Eq _) = true
	      | loop(NotEq _,NotEq _) = true
	      | loop(Not _,Not _) = true
	      | loop(Forall(_,x1,y1),Forall(_,x2,y2)) =
		    (looplist(x1,x2)) andalso (loop(y1,y2))
	      | loop(Exists(_,x1,y1),Exists(_,x2,y2)) =
		    (looplist(x1,x2)) andalso (loop(y1,y2))
	      | loop(Ag _,Ag _) = true
	      | loop(Af _,Af _) = true
	      | loop(Eg _,Eg _) = true
	      | loop(Ef _,Ef _) = true
	      | loop(Ax _,Ax _) = true
	      | loop(Ex _,Ex _) = true
	      | loop(Au _,Au _) = true
	      | loop(Eu _,Eu _) = true
	      | loop(Ar _,Ar _) = true
	      | loop(Er _,Er _) = true
	      | loop(Lt _,Lt _) = true
	      | loop(Gt _,Gt _) = true
	      | loop(Le _,Le _) = true
	      | loop(Ge _,Ge _) = true
	      | loop(Plus _,Plus _) = true
	      | loop(Minus _,Minus _) = true
	      | loop(UMinus _,UMinus _) = true
	      | loop(Times _,Times _) = true
	      | loop(Div _,Div _) = true
	      | loop(Mod _,Mod _) = true
	      | loop(Number(_,n1),Number(_,n2)) = (n1 = n2)
	      | loop(BoolType _,BoolType _) = true
	      | loop(BoolType2,BoolType2) = true
	      | loop(NatType _,NatType _) = true
	      | loop(IntType _,IntType _) = true
	      | loop(RangeType(_,x1,y1),RangeType(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(FunType(_,x1,y1),FunType(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(TupleType(_,lst1),TupleType(_,lst2)) =
		    (looplist(lst1,lst2))
	      | loop(ArrayType(_,x1,y1),ArrayType(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(RecordType(_,lst1),RecordType(_,lst2)) =
		    (looplist(lst1,lst2))
	      | loop(EnumType(_,lst1),EnumType(_,lst2)) =
		    (looplist(lst1,lst2))
	      | loop(AbstractType(lst1),AbstractType(lst2)) =
		    (looplist(lst1,lst2))
	      | loop(Of(_,x1,y1),Of(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(TypedVars(_,lst1,o1),TypedVars(_,lst2,o2)) =
		    (looplist(lst1,lst2)) andalso
		    (case (o1,o2) of
			 (NONE,NONE) => true
		       | (SOME(x1),SOME(x2)) => loop(x1,x2)
		       | _ => false)
	      | loop(False _,False _) = true
	      | loop(True _,True _) = true
	      | loop(False2,False2) = true
	      | loop(True2,True2) = true
	      | loop(Min _,Min _) = true
	      | loop(Max _,Max _) = true
	      | loop(Undefined _,Undefined _) = true
	      | loop(Anyvalue _,Anyvalue _) = true
	      | loop(OtherValue,OtherValue) = true
	      | loop(Self _,Self _) = true
	      | loop(Appl(_,x1,y1),Appl(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(TypedExpr(_,x1,y1),TypedExpr(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(Fn(_,lst1),Fn(_,lst2)) = (looplist(lst1,lst2))
	      | loop(ChoicePattern(_,lst1),ChoicePattern(_,lst2)) =
		    (looplist(lst1,lst2))
	      | loop(ApplPattern(_,x1,y1),ApplPattern(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(AsPattern(_,x1,y1),AsPattern(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(TypedPattern(_,x1,y1),TypedPattern(_,x2,y2)) =
		    (loop(x1,x2)) andalso (loop(y1,y2))
	      | loop(TuplePattern(_,lst1),TuplePattern(_,lst2)) =
		    (looplist(lst1,lst2))
	      | loop(RecordPattern(_,lst1),RecordPattern(_,lst2)) =
		    (looplist(lst1,lst2))
	      | loop(Underscore _,Underscore _) = true
	      | loop(ExtractAppl(y1,z1),ExtractAppl(y2,z2)) =
		    loop(y1,y2) andalso loop(z1,z2)
	      | loop(ExtractTuple(n1,x1),ExtractTuple(n2,x2)) =
		    n1 = n2 andalso loop(x1,x2)
	      | loop(ExtractRecord(x1,y1),ExtractRecord(x2,y2)) =
		    loop(x1,x2) andalso loop(y1,y2)
	      | loop(ExtractIndex x1, ExtractIndex x2) = loop(x1,x2)
	      | loop(Group(_,x1),Group(_,x2)) = loop(x1,x2)
	      | loop(Fake,Fake) = true
	      | loop _ = false
	in loop(t1,t2)
	end

    (* Make the default equality NOT to compare parent closures.  It
       doesn't seem to be necessary, and is too expensive. *)
    val ptEq = ptEqCommon true

    (* Don't compare parent closures, consider objects equivalent if their
     * local definitions are the same. *)
    val ptEqLoose = ptEqCommon true

(* Takes a function `ff' that transforms the parse tree, and applies
   it to every subexpression of `expr' of ParseTree type.  This
   function may call `ptTransform' on its argument, and thus,
   implement the recursive descend for the entire tree without having
   to mention all of the cases.

   Parent closures are not changed.  If you need to recurse into
   parent closures, consider those cases in `ff'. 

   Also, since this is a very critical function, we explicitly
   enumerate all cases in the ParseTree, to let the ML typechecker
   catch anything we've missed. *)

    fun ptTransform ff expr =
	let val fflist = List.map ff
	    fun loop (Program(p,lst)) = Program(p,fflist lst)
	      | loop (x as Include _) = x
	      | loop (Included(p,s,x)) = Included(p,s,ff x)
	      | loop (x as Quote _) = x
	      (* Open the reference *)
	      | loop (SharedRef x) = ff(!x)
	      | loop (x as Id _) = x
	      | loop (x as Uid _) = x
	      | loop (x as Tvar _) = x
	      | loop (x as TvarN _) = x
	      | loop (TypeInst(p,lst,x)) = TypeInst(p,fflist lst, ff x)
	      | loop (x as Ellipsis _) = x
	      | loop (TopClosure lst) = TopClosure(fflist lst)
	      | loop (Module(p,{name=name,
				statparams=stats,
				dynparams=dyn,
				body=body})) =
		  Module(p,{name=name,
			    statparams=Option.map fflist stats,
			    dynparams=Option.map ff dyn,
			    body=ff body})
	      | loop (ModuleClosure{name=name,
				    uname=uname,
				    dynName=dn,
				    Sig=msig,
				    closure=cl,
				    def=def,
				    parent=p}) =
		  ModuleClosure{name=name,
				uname=uname,
				dynName=Option.map ff dn,
				Sig=ff msig,
				closure=fflist cl,
				def=Option.map ff def,
				parent=p}
	      | loop (ModuleSig{statparams=stats}) =
		  ModuleSig{statparams=Option.map fflist stats}
	      | loop (ModuleParamClosure{closure=cl, parent=parent}) =
		  ModuleParamClosure{closure=fflist cl, parent=parent}
	      | loop (FunClosure{name=name,
				 formals=arg,
				 parent=parent,
				 body=body}) =
		  FunClosure{name=name,
			     formals=ff arg,
			     parent=parent,
			     body=ff body}
	      | loop (RecurFun x) = RecurFun(ff x)
	      | loop (LetClosure{locals=locals,
				 parent=parent,
				 body=body}) =
		  LetClosure{locals=fflist locals,
			     parent=parent,
			     body=ff body}
	      | loop (LetAsstClosure{locals=locals,
				     parent=parent,
				     body=body}) =
		  LetAsstClosure{locals=fflist locals,
				 parent=parent,
				 body=fflist body}
	      | loop (TypeClosure{name=name,
				  uname=uname,
				  params=params,
				  def=def,
				  recursive=r,
				  parent=parent}) =
		  TypeClosure{name=name,
			      uname=uname,
			      params=params,
			      def=ff def,
			      recursive=r,
			      parent=parent}
	      | loop (ChoiceClosure{pattern=pat,
				    uname=uname,
				    names=names,
				    body=body,
				    parent=parent}) =
		  ChoiceClosure{pattern=ff pat,
				uname=uname,
				names=fflist names,
				body=ff body,
				parent=parent}
	      | loop (ChoiceAsstClosure{pattern=pat,
					uname=uname,
					names=names,
					body=body,
					parent=parent}) =
		  ChoiceAsstClosure{pattern=ff pat,
				    uname=uname,
				    names=fflist names,
				    body=fflist body,
				    parent=parent}
	      | loop (ChooseClosure{pattern=pat,
				    Type=tp,
				    names=names,
				    body=body,
				    parent=parent}) =
		  ChooseClosure{pattern=ff pat,
				Type=ff tp,
				names=fflist names,
				body=ff body,
				parent=parent}
	      | loop (ForallClosure{names=names,
				    body=body,
				    parent=parent}) =
		  ForallClosure{names=fflist names,
				body=ff body,
				parent=parent}
	      | loop (ExistsClosure{names=names,
				    body=body,
				    parent=parent}) =
		  ExistsClosure{names=fflist names,
				body=ff body,
				parent=parent}
	      | loop (QuantifiedVar{name=n,uname=un,Type=tp}) =
		  QuantifiedVar{name=n, uname=un, Type=ff tp}
	      | loop (SyncClosure{names=names,
				  body=body,
				  parent=parent}) =
		  SyncClosure{names=fflist names,
			      body=ff body,
			      parent=parent}
	      | loop (AsyncClosure{names=names,
				   body=body,
				   parent=parent}) =
		  AsyncClosure{names=fflist names,
			       body=ff body,
			       parent=parent}
	      | loop (Builtin{name=name, Type=tp}) =
		  Builtin{name=name, Type=ff tp}
	      | loop (StateVar{name=n, uname=un, Type=tp, id=id}) =
		  StateVar{name=n, uname=un, Type=ff tp, id=ff id}
	      | loop (Object{name=name,
			     uname=uname,
			     Type=tp,
			     def=def}) =
		  Object{name=name,
			 uname=uname,
			 Type=ff tp,
			 def=ff def}
	      | loop (ObjectInst{obj=obj,subst=f}) =
		  ObjectInst{obj=ff obj, subst=f}
	      | loop (PatternFormal{name=name,
				    uname=uname,
				    Type=tp,
				    extract=extract}) =
		  PatternFormal{name=name,
				uname=uname,
				Type=ff tp,
				extract=extract}
	      | loop (DynPatternFormal{name=name,
				       uname=uname,
				       Type=tp,
				       extract=extract}) =
		  DynPatternFormal{name=name,
				   uname=uname,
				   Type=ff tp,
				   extract=extract}
	      | loop (TypeConstr{name=name,
				 uname=uname,
				 Type=tp}) =
		  TypeConstr{name=name, 
			     uname=uname,
			     Type=ff tp}
	      | loop (RecordField{name=name,
				  Type=tp}) =
		  RecordField{name=name,
			      Type=ff tp}
	      | loop (StaticFormalConst{name=name,
					uname=uname,
					Type=def}) =
		  StaticFormalConst{name=name,
				    uname=uname,
				    Type=ff def}
	      | loop (StaticFormalType{name=name,
				       uname=uname}) =
		  StaticFormalType{name=name,
				    uname=uname}
	      | loop (Theorem(p,name,def)) = Theorem(p,name,ff def)
	      | loop (PType(p,tp)) = PType(p,ff tp)
	      | loop (Symmetric(p,tp)) = Symmetric(p,ff tp)
	      | loop (Finite(p,tp)) = Finite(p,ff tp)
	      | loop (BeginEnd(p,lst)) = BeginEnd(p,fflist lst)
	      | loop (Sync2(p,m1,m2)) = Sync2(p,ff m1,ff m2)
	      | loop (Async2(p,m1,m2)) = Async2(p,ff m1,ff m2)
	      | loop (Sync(p,lst,m)) = Sync(p,fflist lst,ff m)
	      | loop (Async(p,lst,m)) = Async(p,fflist lst,ff m)
	      | loop (TypeDecl(p,name,parms,def)) =
		  TypeDecl(p,name,parms,ff def)
	      | loop (DatatypeDecl(p,name,parms,def)) =
		  DatatypeDecl(p,name,parms,ff def)
	      | loop (Var(p,lst)) = Var(p,fflist lst)
	      | loop (StateVarDecl(p,lst)) = StateVarDecl(p,fflist lst)
	      | loop (Val(p,name,def)) = Val(p,name,ff def)
	      | loop (FunCases(p,lst)) = FunCases(p,fflist lst)
	      | loop (FunCase(p,name,pats,tp,def)) =
		  FunCase(p,name,fflist pats,
			  (case tp of
			       NONE => NONE
			     | SOME(t) => SOME(ff t)), ff def)
	      | loop (Open(p,m)) = Open(p,ff m)
	      | loop (Export(p,lst)) = Export(p,fflist lst)
	      | loop (Models(p,x,y)) = Models(p,ff x,ff y)
	      | loop (Refines(p,x,y)) = Refines(p,ff x,ff y)
	      | loop (x as Nop _) = x
	      | loop (NormalAsst(p,name,e)) = NormalAsst(p,ff name,ff e)
	      | loop (InitAsst(p,name,e)) = InitAsst(p,ff name,ff e)
	      | loop (NextAsst(p,name,e)) = NextAsst(p,ff name,ff e)
	      | loop (Next e) = Next(ff e)
	      | loop (IfAsst(p,lst1,lst2)) = IfAsst(p,fflist lst1,fflist lst2)
	      | loop (CondAsst(p,x,lst)) = CondAsst(p,ff x,fflist lst)
	      | loop (LetAsst(p,defs,lst)) = LetAsst(p,fflist defs,fflist lst)
	      | loop (CaseAsst(p,e,lst)) = CaseAsst(p,ff e,fflist lst)
	      | loop (ChoiceAsst(p,x,lst)) = ChoiceAsst(p,ff x,fflist lst)
	      | loop (ChooseAsst(p,lst)) = ChooseAsst(p,fflist lst)
	      | loop (IfExpr(p,lst,x)) = IfExpr(p,fflist lst,ff x)
	      | loop (CondExpr(p,x,y)) = CondExpr(p,ff x,ff y)
	      | loop (LetExpr(p,defs,e)) = LetExpr(p,fflist defs,ff e)
	      | loop (CaseExpr(p,e,lst)) = CaseExpr(p,ff e,fflist lst)
	      | loop (ChoiceExpr(p,x,y)) = ChoiceExpr(p,ff x,ff y)
	      | loop (ChooseExpr(p,x,y)) = ChooseExpr(p,ff x,ff y)
	      | loop (WithExpr(p,x,lst)) = WithExpr(p,ff x,fflist lst)
	      | loop (WithAsst(p,name,y)) = WithAsst(p,name,ff y)
	      | loop (RecordExpr(p,lst)) = RecordExpr(p,fflist lst)
	      | loop (TupleExpr(p,lst)) = TupleExpr(p,fflist lst)
	      | loop (NondetExpr(p,lst)) = NondetExpr(p,fflist lst)
	      | loop (RecordAsst(p,x,y)) = RecordAsst(p,ff x,ff y)
	      | loop (ModuleInst(p,m,lst)) = ModuleInst(p,ff m,fflist lst)
	      | loop (Dot(p,x,y)) = Dot(p,ff x,ff y)
	      | loop (Type(p,tp)) = Type(p,ff tp)
	      | loop (Datatype(p,tp)) = Datatype(p,tp)
	      | loop (x as And _) = x
	      | loop (x as Or _) = x
	      | loop (x as Implies _) = x
	      | loop (x as Arrow _) = x
	      | loop (x as Darrow _) = x
	      | loop (x as Iff _) = x
	      | loop (x as Eq _) = x
	      | loop (x as NotEq _) = x
	      | loop (x as Not _) = x
	      | loop (Forall(p,lst,x)) = Forall(p,fflist lst,ff x)
	      | loop (Exists(p,lst,x)) = Exists(p,fflist lst,ff x)
	      | loop (x as Ag _) = x
	      | loop (x as Af _) = x
	      | loop (x as Eg _) = x
	      | loop (x as Ef _) = x
	      | loop (x as Ax _) = x
	      | loop (x as Ex _) = x
	      | loop (x as Au _) = x
	      | loop (x as Eu _) = x
	      | loop (x as Ar _) = x
	      | loop (x as Er _) = x
	      | loop (x as Lt _) = x
	      | loop (x as Gt _) = x
	      | loop (x as Le _) = x
	      | loop (x as Ge _) = x
	      | loop (x as Plus _) = x
	      | loop (x as Minus _) = x
	      | loop (x as UMinus _) = x
	      | loop (x as Times _) = x
	      | loop (x as Div _) = x
	      | loop (x as Mod _) = x
	      | loop (x as Number _) = x
	      | loop (x as BoolType _) = x
	      | loop (x as BoolType2) = x
	      | loop (x as NatType _) = x
	      | loop (x as IntType _) = x
	      | loop (RangeType(p,x,y)) = RangeType(p,ff x,ff y)
	      | loop (FunType(p,x,y)) = FunType(p,ff x,ff y)
	      | loop (TupleType(p,x)) = TupleType(p,fflist x)
	      | loop (ArrayType(p,x,y)) = ArrayType(p,ff x,ff y)
	      | loop (RecordType(p,x)) = RecordType(p,fflist x)
	      | loop (EnumType(p,lst)) = EnumType(p,fflist lst)
	      | loop (AbstractType(lst)) = AbstractType(fflist lst)
	      | loop (Of(p,name,tp)) = Of(p,name,ff tp)
	      | loop (TypedVars(p,lst,tp)) = TypedVars(p,fflist lst, Option.map ff tp)
	      | loop (x as False _) = x
	      | loop (x as True _) = x
	      | loop (x as False2) = x
	      | loop (x as True2) = x
	      | loop (x as Min _) = x
	      | loop (x as Max _) = x
	      | loop (x as Undefined _) = x
	      | loop (x as Anyvalue _) = x
	      | loop (x as OtherValue) = x
	      | loop (x as Self _) = x
	      | loop (Appl(p,x,y)) = Appl(p,ff x,ff y)
	      | loop (TypedExpr(p,x,y)) = TypedExpr(p,ff x,ff y)
	      | loop (Fn(p,lst)) = Fn(p,fflist lst)
	      | loop (ChoicePattern(p,lst)) = ChoicePattern(p,fflist lst)
	      | loop (ApplPattern(p,x,y)) = ApplPattern(p,ff x,ff y)
	      | loop (TypedPattern(p,x,y)) = TypedPattern(p,ff x,ff y)
	      | loop (TuplePattern(p,lst)) = TuplePattern(p,fflist lst)
	      | loop (RecordPattern(p,lst)) = RecordPattern(p,fflist lst)
	      | loop (AsPattern(p,x,y)) = AsPattern(p,ff x,ff y)
	      | loop (x as Underscore _) = x
	      | loop (ExtractAppl(y,z)) = ExtractAppl(ff y,ff z)
	      | loop (ExtractTuple(n,x)) = ExtractTuple(n,ff x)
	      | loop (ExtractRecord(x,y)) = ExtractRecord(ff x,ff y)
	      | loop (ExtractIndex x) = ExtractIndex(ff x)
	      | loop (Group(p,x)) = Group(p,ff x)
	      | loop (x as Fake) = x
	in loop expr
	end

    (* Same as ptTransformParent, only ff takes context and ParseTree
     and updates the parent closures correctly, so that they point to
     the closures of the current tree. 

     If the initial context is "Fake", then the outermost context is
     kept unchanged.  That is, if you want the expression to remain in
     the context it comes from, simply give "Fake" as the current
     context. *)

    fun ptTransformParent ff context expr =
	let fun fflist context lst = List.map (ff context) lst
	    fun pickC(parent,context) =
		  (case context of
		       Fake => parent
		     | _ => context)
	    fun descend c [] acc = acc
	      | descend c (d::lst) acc = descend c lst ((ff (c acc) d)::acc)
	    fun recur c x = ptTransform (ff c) x
	    (* Attention: Module doesn't form a context, only
	       ModuleClosure does, so this case should be taken care
	       externally if needed. *)
	    fun loop context (Module(p,{name=name,
					statparams=stats,
					dynparams=dyn,
					body=body})) =
		let val sp = Option.map(fflist context) stats
		    val dp = Option.map(ff context) dyn
		in Module(p,{name=name,
			     statparams=sp,
			     dynparams=dp,
			     body=ff context body})
		end
	      | loop context (ModuleClosure{name=name,
					    uname=uname,
					    dynName=dn,
					    Sig=msig,
					    closure=cl,
					    def=def,
					    parent=p}) =
		let val newdef = Option.map (ff context) def
		    val newSig=ff context msig
		    fun c lst = 
			 ModuleClosure{name=name,
				       uname=uname,
				       dynName=dn,
				       Sig=newSig,
				       closure=lst,
				       def=newdef,
				       parent=context}
		    val cl = descend c (List.rev cl) []
		in (c cl)
		end
	      | loop context (ModuleParamClosure{closure=cl, parent=p}) =
		let fun c lst = ModuleParamClosure{closure=lst, 
						 parent=pickC(p,context)}
		    val cl = descend c (List.rev cl) []
		in c cl
		end
	      | loop context (FunClosure{name=name,
				 formals=arg,
				 parent=p,
				 body=body}) =
		let val cxt = pickC(p,context)
		    val arg = ff context arg
		    val b = ff (FunClosure{name=name,
					   formals=arg,
					   parent=cxt,
					   body=Fake}) body
		in FunClosure{name=name,
			      formals=arg,
			      parent=cxt,
			      body=b}
		end
	      | loop context (LetClosure{locals=locals,
				 parent=p,
				 body=body}) =
		let fun c b lst = 
		      LetClosure{locals=lst,
				 parent=pickC(p,context),
				 body=b}
		    val locs = descend (c Fake) (List.rev locals) []
		in c (ff (c Fake locs) body) locs
		end
	      | loop context (LetAsstClosure{locals=locals,
					     parent=p,
					     body=body}) =
		let fun c b lst = 
		      LetAsstClosure{locals=lst,
				     parent=pickC(p,context),
				     body=b}
		    val locs = descend (c []) (List.rev locals) []
		    val cxt = c [] locs
		in c (fflist cxt body) locs
		end
	      | loop context (TypeClosure{name=name,
				  uname=uname,
				  params=params,
				  def=def,
				  recursive=r,
				  parent=p}) =
		  TypeClosure{name=name,
			      uname=uname,
			      params=params,
			      def=ff (pickC(p,context)) def,
			      recursive=r,
			      parent=(pickC(p,context))}
	      | loop context (ChoiceClosure{pattern=pat,
					    uname=uname,
					    names=names,
					    body=body,
					    parent=pc}) =
		  let val cxt=pickC(pc,context)
		      val newnames = fflist cxt names
		      fun c b pat =
		        ChoiceClosure{pattern=pat,
				      uname=uname,
				      names=newnames,
				      body=b,
				      parent=cxt}
		      val pat = ff cxt pat
		  in c (ff (c Fake pat) body) pat
		  end
	      | loop context (ChoiceAsstClosure{pattern=pat,
						uname=uname,
						names=names,
						body=body,
						parent=pc}) =
		  let val cxt=pickC(pc,context)
		      val newnames = fflist cxt names
		      fun c b pat =
		        ChoiceAsstClosure{pattern=pat,
					  uname=uname,
					  names=newnames,
					  body=b,
					  parent=cxt}
		      val pat = ff cxt pat
		  in c (fflist (c [] pat) body) pat
		  end
	      | loop context (ChooseClosure{pattern=pat,
					    Type=tp,
					    names=names,
					    body=body,
					    parent=pc}) =
		  let val cxt=pickC(pc,context)
		      val newnames = fflist cxt names
		      val pat = ff cxt pat
		      val tp = ff cxt tp
		      fun c b =
		        ChooseClosure{pattern=pat,
				      Type=tp,
				      names=newnames,
				      body=b,
				      parent=cxt}
		  in c (ff (c Fake) body)
		  end
	      | loop context (ForallClosure{names=names,
				    body=body,
				    parent=pc}) =
		  let val cxt=pickC(pc,context)
		      fun c b n =
		        ForallClosure{names=n,
				      body=b,
				      parent=cxt}
		      val names=fflist cxt names
		  in c (ff (c Fake names) body) names
		  end
	      | loop context (ExistsClosure{names=names,
				    body=body,
				    parent=pc}) =
		  let val cxt=pickC(pc,context)
		      fun c b n =
		        ExistsClosure{names=n,
				      body=b,
				      parent=cxt}
		      val names=fflist cxt names
		  in c (ff (c Fake names) body) names
		  end
	      | loop context (StateVar{name=n, uname=un, Type=tp, id=id}) =
		  StateVar{name=n,
			   uname=un,
			   Type=ff context tp,
			   id= ff context id}
	      | loop context (SyncClosure{names=names,
					  body=body,
					  parent=pc}) =
		  let val cxt=pickC(pc,context)
		      fun c b n =
			  SyncClosure{names=n,
				      body=b,
				      parent=cxt}
		      val names=fflist cxt names
		  in c (ff (c Fake names) body) names
		  end
	      | loop context (AsyncClosure{names=names,
					   body=body,
					   parent=pc}) =
		  let val cxt=pickC(pc,context)
		      fun c b n =
			  AsyncClosure{names=n,
				       body=b,
				       parent=cxt}
		      val names=fflist cxt names
		  in c (ff (c Fake names) body) names
		  end
	      | loop context (TopClosure lst) =
		  let fun c lst = TopClosure lst
		  in TopClosure(descend c (List.rev lst) [])
		  end
	      | loop context x = recur context x
	in loop context expr
	end

    (* transforms in a bottom-up manner *)


    fun ptTransformUp ff expr =
      ff (ptTransform (ptTransformUp ff) expr)

    (* Define what is considered a name, in one row with Id *)
    fun isName (Id _) = true
      | isName (And _) = true
      | isName (Or _) = true
      | isName (Implies _) = true
      | isName (Darrow _) = true
      | isName (Arrow _) = true
      | isName (Iff _) = true
      | isName (Eq _) = true
      | isName (NotEq _) = true
      | isName (Not _) = true
      | isName (Ag _) = true
      | isName (Eg _) = true
      | isName (Af _) = true
      | isName (Ef _) = true
      | isName (Ax _) = true
      | isName (Ex _) = true
      | isName (Au _) = true
      | isName (Eu _) = true
      | isName (Ar _) = true
      | isName (Er _) = true
      | isName (Lt _) = true
      | isName (Gt _) = true
      | isName (Le _) = true
      | isName (Ge _) = true
      | isName (Plus _) = true
      | isName (Minus _) = true
      | isName (UMinus _) = true
      | isName (Times _) = true
      | isName (Div _) = true
      | isName (Mod _) = true
      | isName (Anyvalue _) = true
      | isName (Undefined _) = true
      | isName (True _) = true
      | isName (False _) = true
      | isName (Self _) = true
      | isName _ = false

    (* Extracts the parent closure from the given one, if one exists.
     * Returns SOME(parent_closure) on success, or NONE on failure. *)
    fun parentClosure (ModuleClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (ModuleParamClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (FunClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (LetClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (LetAsstClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (TypeClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (ChoiceClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (ChoiceAsstClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (ChooseClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (ForallClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (ExistsClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (SyncClosure{parent=cl,...}) = SOME(cl)
      | parentClosure (AsyncClosure{parent=cl,...}) = SOME(cl)
      | parentClosure _ = NONE

    (* Extract the type field from any object. Return NONE or SOME tp. *)
    fun GetObjectType (Object{Type=tp,...}) = SOME tp
      | GetObjectType (ObjectInst{obj=obj,subst=ff}) = Option.map ff (GetObjectType obj)
      | GetObjectType (QuantifiedVar{Type=tp,...}) = SOME tp
      | GetObjectType (StateVar{Type=tp,...}) = SOME tp
      | GetObjectType (PatternFormal{Type=tp,...}) = SOME tp
      | GetObjectType (DynPatternFormal{Type=tp,...}) = SOME tp
      | GetObjectType (TypeConstr{Type=tp,...}) = SOME tp
      | GetObjectType (RecordField{Type=tp,...}) = SOME tp
      | GetObjectType (StaticFormalConst{Type=tp,...}) = SOME tp
      | GetObjectType (Type(_,tp)) = SOME tp
      | GetObjectType (Datatype(_,tp)) = SOME tp
      | GetObjectType _ = NONE

    (* Extract the name field from any object *)
    fun GetObjectName (Object{name=n,...}) = SOME n
      | GetObjectName (ObjectInst{obj=obj,...}) = GetObjectName obj
      | GetObjectName (QuantifiedVar{name=n,...}) = SOME n
      | GetObjectName (StateVar{name=n,...}) = SOME n
      | GetObjectName (PatternFormal{name=n,...}) = SOME n
      | GetObjectName (DynPatternFormal{name=n,...}) = SOME n
      | GetObjectName (TypeConstr{name=n,...}) = SOME n
      | GetObjectName (RecordField{name=n,...}) = SOME n
      | GetObjectName (StaticFormalConst{name=n,...}) = SOME n
      | GetObjectName (StaticFormalType{name=n,...}) = SOME n
      | GetObjectName (Module(_,{name=n,...})) = SOME n
      | GetObjectName (ModuleClosure{name=no,...}) = no
      | GetObjectName (TypeClosure{name=n,...}) = SOME n
      | GetObjectName (PType(_,n)) = SOME n
      | GetObjectName (FunClosure{name=no,...}) = no
      | GetObjectName (Theorem(_,n,_)) = SOME n
      | GetObjectName x = if isName x then SOME x else NONE

    (* Extract the uname field from any object *)
    fun GetObjectUName (Object{uname=n,...}) = SOME n
      | GetObjectUName (ObjectInst{obj=obj,...}) = GetObjectUName obj
      | GetObjectUName (QuantifiedVar{uname=n,...}) = SOME n
      | GetObjectUName (StateVar{uname=n,...}) = SOME n
      | GetObjectUName (PatternFormal{uname=n,...}) = SOME n
      | GetObjectUName (DynPatternFormal{uname=n,...}) = SOME n
      | GetObjectUName (TypeConstr{uname=n,...}) = SOME n
      | GetObjectUName (StaticFormalConst{uname=n,...}) = SOME n
      | GetObjectUName (StaticFormalType{uname=n,...}) = SOME n
      | GetObjectUName (ModuleClosure{uname=n,...}) = SOME n
      | GetObjectUName (TypeClosure{uname=n,...}) = SOME n
      | GetObjectUName (n as Uid _) = SOME n
      | GetObjectUName x = NONE
      
    (* Determine if a ParseTree represents a constant value or a non-value. *)

    fun isValue parseTree = 
      case parseTree of
	FunClosure {body = x,...} => isValue x
      | RecurFun (F) => isValue F
      | Builtin{name=Anyvalue _,...} => false
      | Builtin _ => true
      | Object {def=x,...} => isValue (x)
      | Appl (_, TypeConstr _,y) => isValue y
      | TypeConstr _ => true
      | Number(_) => true
      | True(x) => true
      | False(x) => true
      | StaticFormalConst(x) => false
      | PatternFormal(x) => false
      | DynPatternFormal(x) => false
      | RecordExpr (rpos, rlist) => List.all 
	    (fn RecordAsst(_,_,PT) => 
	     isValue PT
	  | _ => raise SympParseInternalError 
	     ("Badly formed record: " 
	      ^ (pt2stringDebug parseTree))) rlist
      | TupleExpr (rpos, tlist) => List.all isValue tlist
      | ObjectInst {obj = x,...} => isValue x
      | _ => false

    (* Determine if a ParseTree represents an undefined value.  Note
       that tuples and records are considered undefined if they
       contain an undefined field. *)

    fun isUndefined (Undefined _) = true
      | isUndefined (Builtin{name=Undefined _,...}) = true
      | isUndefined (RecordExpr(_,lst)) = List.exists isUndefined lst
      | isUndefined (RecordAsst(_,_,x)) = isUndefined x
      | isUndefined (TupleExpr(_,lst)) = List.exists isUndefined lst
      | isUndefined (ExtractRecord(_,x)) = isUndefined x
      | isUndefined (ExtractTuple(_,x)) = isUndefined x
      | isUndefined (ExtractAppl(_,x)) = isUndefined x
      | isUndefined (ExtractIndex x) = isUndefined x
      | isUndefined (Object{def=x,...}) = isUndefined x
      | isUndefined (ObjectInst{obj=x,...}) = isUndefined x
      | isUndefined _ = false

    (* Strip object down to its component values, without names. *)

    fun stripObjects (Object{def=x,...}) = stripObjects (x)
      | stripObjects (TupleExpr(tpos,tlist)) = TupleExpr (tpos,
							List.map 
							stripObjects tlist)
      | stripObjects (re as RecordExpr (rpos, rlist)) = 
      RecordExpr (rpos, List.map (fn RecordAsst(a, b, PT) => 
				  RecordAsst (a, b, stripObjects PT)
    | _ => raise SympParseInternalError 
				  ("Badly formed record: " 
				   ^ (pt2stringDebug re))) rlist)
      | stripObjects (Appl (p, x, y)) = Appl (p, stripObjects x, 
					      stripObjects y)
      | stripObjects x = x

    (* Strip all ObjectInst's from the expression, instantiating the types properly. *)
    fun stripObjectInst expr =
      let fun looplist ff cxt lst = List.map (loop ff cxt) lst
	  and recur ff context x = ptTransformParent (loop ff) context  x
	  and loop ff cxt (ObjectInst{obj=x,subst=subst}) = loop (ff o subst) cxt x
	    | loop ff cxt (TypeInst(p,parms,tp)) = TypeInst(p,List.map ff parms, ff tp)
	    | loop ff cxt (ChooseClosure{pattern=pat,
					 Type=tp,
					 names=names,
					 body=b,
					 parent=p}) =
	        let val newNames = looplist ff cxt names
		    val newPattern = loop ff cxt pat
		    fun cc body =
			ChooseClosure{pattern=newPattern,
				      Type=ff tp,
				      names=newNames,
				      body=body,
				      parent=cxt}
		    val body = loop ff (cc Fake) b
		in cc body
		end
	    | loop ff cxt (QuantifiedVar{name=n,uname=uname,Type=tp}) =
		QuantifiedVar{name=n,uname=uname,Type=ff tp}
	    | loop ff cxt (StateVar{name=name,uname=uname,Type=tp,id=id}) =
		StateVar{name=name,uname=uname,Type=ff tp,id=loop ff cxt id}
	    | loop ff cxt (Object{name=name,uname=uname,Type=tp,def=def}) =
		Object{name=name,uname=uname,Type=ff tp,def=loop ff cxt def}
	    | loop ff cxt (PatternFormal{name=name,uname=uname,Type=tp,extract=extract}) =
		PatternFormal{name=name,uname=uname,Type=ff tp,extract=extract}
	    | loop ff cxt (DynPatternFormal{name=name,uname=uname,Type=tp,extract=extract}) =
		DynPatternFormal{name=name,uname=uname,Type=ff tp,extract=extract}
	    | loop ff cxt (TypeConstr{name=name,uname=uname,Type=tp}) =
		TypeConstr{name=name,uname=uname,Type=ff tp}
	    | loop ff cxt (RecordField{name=name,Type=tp}) =
		RecordField{name=name,Type=ff tp}
	    | loop ff cxt (StaticFormalConst{name=name,uname=uname,Type=tp}) =
		StaticFormalConst{name=name,uname=uname,Type=ff tp}
	    | loop ff cxt (PType(p,tp)) = PType(p,ff tp)
	    | loop ff cxt (Symmetric(p,tp)) = Symmetric(p,ff tp)
	    | loop ff cxt (Finite(p,tp)) = Finite(p,ff tp)
	    | loop ff cxt (TypedExpr(p,e,tp)) = TypedExpr(p,loop ff cxt e,ff tp)
	    | loop ff cxt (TypedPattern(p,e,tp)) = TypedPattern(p,loop ff cxt e,ff tp)
	    | loop ff cxt x = recur ff cxt x
      in loop (fn x=>x) Fake expr
      end

    fun eqUNames (e1, e2) =
      case (GetObjectUName e1) of
	SOME u1 =>
	  (case (GetObjectUName e2) of
	     SOME u2 =>
	       ptEq (u1, u2)
	   | NONE => false)
      | NONE => false

    (* Additional functions to extract a type of any typechecked expression. *)
    fun getExprType findObject expr =
	let fun loop ff (ObjectInst{obj=obj,subst=subst}) =
	         loop (ff o subst) obj
	      | loop ff (Object{Type=tp,...}) = ff tp
	      | loop ff (Next e) = loop ff e
	      | loop ff (FunClosure{formals=arg,body=body,...}) =
		 FunType(dp,loop ff arg, loop ff body)
	      | loop ff (RecurFun F) =
		 (case loop ff F of
		      FunType(_,_,tp) => tp
		    | tp => raise SympBug("getExprType: RecurFun is not FunType: "
					  ^(pt2stringDebug tp)))
	      | loop ff (LetClosure{body=body,...}) = loop ff body
	      | loop ff (ChoiceClosure{body=body,...}) = loop ff body
	      | loop ff (ChooseClosure{Type=tp,...}) = ff tp
	      | loop ff (ForallClosure _) = BoolType dp
	      | loop ff (ExistsClosure _) = BoolType dp
	      | loop ff (QuantifiedVar{Type=tp,...}) = ff tp
	      | loop ff (Builtin{Type=tp,...}) = ff tp
	      | loop ff (Number _) = IntType dp
	      | loop ff (True _) = BoolType dp
	      | loop ff (False _) = BoolType dp
	      | loop ff True2 = BoolType2
	      | loop ff False2 = BoolType2
	      | loop ff (StateVar{Type=tp,...}) = ff tp
	      | loop ff (PatternFormal{Type=tp,...}) = ff tp
	      | loop ff (DynPatternFormal{Type=tp,...}) = ff tp
	      | loop ff (TypeConstr{Type=tp,...}) = ff tp
	      | loop ff (RecordField{Type=tp,...}) = ff tp
	      | loop ff (StaticFormalConst{Type=tp,...}) = ff tp
	      | loop ff (e as (IfExpr(_,[],_))) = raise SympBug
		      ("getExprType: IfExpr has empty `if' list: "
		       ^(pt2stringDebug e))
	      | loop ff (IfExpr(_,hd::_,_)) = loop ff hd
	      | loop ff (CondExpr(_,_,e)) = loop ff e
	      | loop ff (LetExpr(_,_,e)) = loop ff e
	      | loop ff (e as CaseExpr(_,_,[])) = raise SympBug
		      ("getExprType: CaseExpr has empty list of cases: "
		       ^(pt2stringDebug e))
	      | loop ff (CaseExpr(_,_,e::_)) = loop ff e
	      | loop ff (WithExpr(_,e,_)) = loop ff e
	      | loop ff (RecordExpr(_,lst)) = RecordType(dp,List.map(loop ff) lst)
	      | loop ff (RecordAsst(_,field,e)) = RecordField{name=field,Type=loop ff e}
	      | loop ff (TupleExpr(_,lst)) = TupleType(dp,List.map(loop ff) lst)
	      | loop ff (NondetExpr(_,e::_)) = loop ff e
	      | loop ff (NondetExpr(p,[])) = raise SympBug
		      ("getExprType: empty NondetExpr at "^(pos2string p))
	      | loop ff (ExtractAppl(c,_)) =
		  (case loop ff c of
		       FunType(_,tp,_) => ff tp
		     | tp => raise SympBug
			   ("getExprType: type constructor in ExtractAppl is not FunType: "
			    ^(pt2stringDebug tp)))
	      | loop ff (ExtractTuple(n,e)) =
		  (case loop ff e of
		       TupleType(_,lst) =>
			   if (List.length lst) > n then List.nth(lst,n)
			   else raise SympBug
			       ("getExprType: expr in ExtractTuple is a shorter tuple than n")
		     | x => raise SympBug
			       ("getExprType: expr in ExtractTuple is not TupleType"))
	      | loop ff (ExtractRecord(f,e)) =
		  let fun findField ((RecordField{name=name,Type=tp})::lst) = 
		           if ptEq(f,name) then ff tp else (findField lst)
			| findField (x::_) = raise SympBug
			   ("getExprType: not RecordField: "^(pt2stringDebug x))
			| findField [] = raise SympBug
			   ("getExprType: no such field in a record: "
			    ^(pt2stringDebug f))
		  in (case loop ff e of
			  RecordType(_,lst) => findField lst
			| x => raise SympBug
			   ("getExprType: expr in ExtractRecord is not RecordType: "
			    ^(pt2stringDebug x)))
		  end
	      (* The type of index is an AbstractType that lists all TypeConstr's. *)
	      | loop ff (ExtractIndex e) =
		  let fun getClist (EnumType(_,lst)) = lst
			| getClist (TypeClosure{def=EnumType(_,lst),...}) = lst
			| getClist (u as Uid _) = 
		           (case findObject u of
				SOME tp => getClist tp
			      | NONE => raise SympBug
				  ("getExprType: no object with such Uid: "
				   ^(pt2string u)))
			| getClist t = raise SympBug
			   ("getExprType: not an EnumType expression in ExtractIndex:\n  "
			    ^(pt2string(TypedExpr(dp,e,t))))
		      val lst = getClist(loop ff e)
		  in AbstractType lst
		  end
	      | loop ff (Appl(_,e,_)) =
		  (case loop ff e of
		       FunType(_,_,tp) => ff tp
		     | tp => raise SympBug("getExprType: function in Appl is not FunType: "
					   ^(pt2stringDebug tp)))
	      | loop ff x = raise SympBug("getExprType: unexpected expression: "
					  ^(pt2stringDebug x))
	in loop (fn x=>x) expr
	end

    (* Checks whether the expression might have types in it.  If it
       does, we have to keep Object and ObjectInst wrappers. *)
    fun hasTypes e =
	(case e of
	     FunClosure _ => true
	   | RecurFun _ => true
(* We cannot wrap these back into Object/ObjectInst, since we lose the
   ability to evaluate them uniformly.  But if we don't, we lose the
   type information.  Maybe it's a tolerable loss, we'll see. *)
(*  	   | Builtin(Undefined _) => true *)
(*  	   | Builtin(Eq _) => true *)
	   | Object _ => true
	   | ObjectInst _ => true
	   | RecordExpr(_,rlist) => List.exists hasTypes rlist
	   | RecordAsst(_,_,x) => hasTypes x
	   | TupleExpr(_,lst) => List.exists hasTypes lst
	   | x => not(isValue x))

    (* Create an `undefined' value of type `tp' *)
    fun bot tp = Builtin{name=Undefined dp, Type=tp}

    (* Try to extract a function value from f.  Also, if the function
       is wrapped into ObjectInst, return a function that will wrap
       any parse tree into the same ObjectInst.  Later, when the
       extracted function is applied, the result will be wrapped this
       way to preserve the type information. *)

    fun extractFun ff f =
	let fun loop ff hh (f as FunClosure _) = SOME(f,hh)
	      | loop ff hh (RecurFun(F)) = SOME(RecurFun(F),hh)
	      | loop ff hh (oi as ObjectInst {obj = x,subst=subst}) =
		  loop (ff o subst) (fn e=>hh(ObjectInst{obj=e,subst=subst})) x
	      | loop ff hh (Object {def = x,...}) = loop ff hh x
	      (* Here we assume the typechecker guarantees they _are_ functions *)
	      | loop ff hh (Builtin{name=f,Type=tp}) = SOME(Builtin{name=f,Type=ff tp},hh)
	      | loop ff hh (TypeConstr{name=name,uname=uname,Type=tp}) =
		  SOME(TypeConstr{name=name,uname=uname,Type=ff tp},hh)
	      | loop _ _ _ = NONE
	in loop ff (fn x=>x) f
	end

    fun extractTuple n (t as TupleExpr(_,l)) = 
	  if List.length l > n then SOME(List.nth(l,n))
	  else raise SympBug("Evaluator/extractTuple: tuple is shorter than "
			     ^(Int.toString n)^": "
			     ^(pt2string t))
      | extractTuple n (oi as ObjectInst{obj = x,subst=subst}) = 
	  let val res = extractTuple n x
	  in 
	     Option.map(fn e=>if hasTypes e then ObjectInst{obj=e,subst=subst}
			      else e) res
	  end
      | extractTuple n (Object{def=x,...}) = extractTuple n x
      | extractTuple _ x = if isUndefined x then SOME(Undefined dp)
			   else NONE

    fun extractRecord field (r as RecordExpr(_,rlist)) = 
	  let val fieldMatch = List.find (fn RecordAsst (_, fname, e) =>
					       ptEq (fname, field)
		 | x => raise SympBug
					       ("Badly formed record field: " ^
						pt2stringDebug x)) rlist
		 in
		   (case fieldMatch of
		      SOME (RecordAsst(_, _, e)) => SOME e
		    | SOME x => raise SympBug ("This can't happen!")
		    | NONE => raise SympBug ("Field extraction failed: " ^
					       pt2stringDebug(ExtractRecord (field, r))))
		 end
      | extractRecord f (w as WithExpr (_, _, wlist)) =
	  let val fieldMatch = List.find (fn WithAsst (_, fname, e) =>
					  ptEq (fname, f)
	| x => raise SympBug
					  ("Badly formed with field: " ^
					   pt2stringDebug x)) wlist
	  in
	    (case fieldMatch of
	       SOME (WithAsst (_, _, e)) => SOME e
	       | SOME x => raise SympBug ("This can't happen!")
	       | NONE => if isUndefined w then SOME(Undefined dp)
			 else NONE)
	  end
      | extractRecord f (oi as ObjectInst{obj = x,subst=subst}) = 
	  let val res = extractRecord f x
	  in 
	     Option.map(fn e=>if hasTypes e then ObjectInst{obj=e,subst=subst}
			      else e) res
	  end
      | extractRecord f (Object{def=x,...}) = extractRecord f x
      | extractRecord _ x = if isUndefined x then SOME(Undefined dp)
			    else NONE

    fun extractAppl c1 (Appl (_, c2, x)) =
      if eqUNames (c1, c2) then
	SOME (x)
      else
	raise SympBug "Incorrect application extraction"
      | extractAppl _ x = if isUndefined x then SOME(Undefined dp)
			  else NONE

    (* newName(): Generate new unique type variable names (string).
     * Not for the user, the names are "illegal" for the parser. *)
    val (newNumber,
	 nextNumber,
	 resetNumber) =
	let val num = ref 0
	in 
	    (fn () => (!num before (num := (!num) + 1)),
	     fn () => (!num),
	     fn () => (num:=0))
	end

    fun newName () = "X!"^(Int.toString(newNumber()))



    (* Similar to the above, only ff returns a function that,
       depending on the argument, will return the new subtree, thus,
       deferring the final computation of the tree.  We propagate this
       to the rest of the tree and return a function that will later
       reconstruct the new tree.  Note, that the tree is actually
       being traversed, not delayed by wrapping into a function, so
       all side effects from ff will take place. *)

(*      fun ptTransformLazy ff expr = *)
(*  	let fun fflist lst = *)
(*  	          let val res = List.map ff lst *)
(*  		  in fn a =>List.map(fn f=>f a) res *)
(*  		  end *)
(*  	    fun loop (Program(p,lst)) = *)
(*  		  let val res = fflist lst *)
(*  		  in fn a=>Program(p,res a) *)
(*  		  end *)
(*  	      | loop (Included(p,s,x)) =  *)
(*  		  let val res = ff x *)
(*  		  in fn a=>Included(p,s,res a) *)
(*  		  end *)
(*  	      | loop (TypeInst(p,lst,x)) = *)
(*  		  let val l1=fflist lst *)
(*  		      val x1=ff x *)
(*  		  in fn a=>TypeInst(p,l1 a, x1 a) *)
(*  		  end *)
(*  	      | loop (TopClosure lst) = *)
(*  		  let val l1=fflist lst *)
(*  		  in fn a=>TopClosure(l1 a) *)
(*  		  end *)
(*  	      | loop (Module(p,{name=name, *)
(*  				statparams=stats, *)
(*  				dynparams=dyn, *)
(*  				body=body, *)
(*  				closure=cl, *)
(*  				parent=parent})) = *)
(*  		  let val sp=(case stats of *)
(*  				  NONE => NONE *)
(*  				| SOME(lst) => SOME(fflist lst)) *)
(*  		      val d=(case dyn of *)
(*  				 NONE => NONE *)
(*  			       | SOME(x) => SOME(ff x)) *)
(*  		      val b=ff body *)
(*  		      val c=fflist cl *)
(*  		      val pc=ff parent *)
(*  		  in fn a=>Module(p,{name=name, *)
(*  				     statparams=(case sp of *)
(*  						     NONE => NONE *)
(*  						   | SOME x => SOME(x a)), *)
(*  				     dynparams=(case d of *)
(*  						    NONE => NONE *)
(*  						  | SOME x => SOME(x a)), *)
(*  				     body=b a, *)
(*  				     closure=c a, *)
(*  				     parent=pc a}) *)
(*  		  end *)
(*  	      | loop (ModuleSig{statparams=stats}) = *)
(*  		  let val s=(case stats of *)
(*  				 NONE => NONE *)
(*  			       | SOME(lst) => SOME(fflist lst)) *)
(*  		  in fn a=>ModuleSig{statparams= *)
(*  				     (case s of *)
(*  					  NONE => NONE *)
(*  					| SOME x => SOME(x a))} *)
(*  		  end *)
(*  	      | loop (ModuleParamClosure{closure=cl, parent=parent}) = *)
(*  		  let val c=fflist cl *)
(*  		      val p=ff parent *)
(*  		  in fn a=>ModuleParamClosure{closure=c a, parent=p a} *)
(*  		  end *)
(*  	      | loop (FunClosure{name=name, *)
(*  				 formals=arg, *)
(*  				 parent=parent, *)
(*  				 body=body}) = *)
(*  		  let val f=ff arg *)
(*  		      val pc=ff parent *)
(*  		      val b=ff body *)
(*  		  in fn a=>FunClosure{name=name, *)
(*  				      formals=f a, *)
(*  				      parent=pc a, *)
(*  				      body=b a} *)
(*  		  end *)
(*  	      | loop (RecurFun x) =  *)
(*  		  let val y = ff x *)
(*  		  in fn a=>RecurFun(y a) *)
(*  		  end *)
(*  	      | loop (LetClosure{locals=locals, *)
(*  				 parent=parent, *)
(*  				 body=body}) = *)
(*  		  let val l=fflist locals *)
(*  		      val pc=ff parent *)
(*  		      val b=ff body *)
(*  		  in fn a=>LetClosure{locals=l a, *)
(*  				      parent=pc a, *)
(*  				      body=b a} *)
(*  		  end *)
(*  	      | loop (TypeClosure{name=name, *)
(*  				  params=params, *)
(*  				  def=def, *)
(*  				  recursive=r, *)
(*  				  parent=parent}) = *)
(*  		  let val d=ff def *)
(*  		      val pc=ff parent *)
(*  		  in fn a=>TypeClosure{name=name, *)
(*  				       params=params, *)
(*  				       def=d a, *)
(*  				       recursive=r, *)
(*  				       parent=pc a} *)
(*  		  end *)
(*  	      | loop (ChoiceClosure{pattern=pat, *)
(*  				    names=names, *)
(*  				    body=body, *)
(*  				    parent=parent}) = *)
(*  		  let val p=ff pat *)
(*  		      val n=fflist names *)
(*  		      val b=ff body *)
(*  		      val pc=ff parent *)
(*  		  in fn a=>ChoiceClosure{pattern=p a, *)
(*  					 names=n a, *)
(*  					 body=b a, *)
(*  					 parent=pc a} *)
(*  		  end *)
(*  	      | loop (ChooseClosure{pattern=pat, *)
(*  				    names=names, *)
(*  				    body=body, *)
(*  				    parent=parent}) = *)
(*  		  let val p=ff pat *)
(*  		      val n=fflist names *)
(*  		      val b=ff body *)
(*  		      val pc=ff parent *)
(*  		  in fn a=>ChooseClosure{pattern=p a, *)
(*  					 names=n a, *)
(*  					 body=b a, *)
(*  					 parent=pc a} *)
(*  		  end *)
(*  	      | loop (ForallClosure{names=names, *)
(*  				    body=body, *)
(*  				    parent=parent}) = *)
(*  		  let val n=fflist names *)
(*  		      val b=ff body *)
(*  		      val pc=ff parent *)
(*  		  in fn a=>ForallClosure{names=n a, *)
(*  					 body=b a, *)
(*  					 parent=pc a} *)
(*  		  end *)
(*  	      | loop (ExistsClosure{names=names, *)
(*  				    body=body, *)
(*  				    parent=parent}) = *)
(*  		  let val n=fflist names *)
(*  		      val b=ff body *)
(*  		      val pc=ff parent *)
(*  		  in fn a=>ExistsClosure{names=n a, *)
(*  					 body=b a, *)
(*  					 parent=pc a} *)
(*  		  end *)
(*  	      | loop (QuantifiedVar{name=n,Type=tp}) = *)
(*  		  let val t=ff tp *)
(*  		  in fn a=>QuantifiedVar{name=n, Type=t a} *)
(*  		  end *)
(*  	      | loop (StateVar{name=n, Type=tp, id=id, context=cxt}) = *)
(*  		  let val t=ff tp *)
(*  		      val id=ff id *)
(*  		      val cxt=ff cxt *)
(*  		  in fn a=>StateVar{name=n, Type=t a, id=id a, context=cxt a} *)
(*  		  end *)
(*  	      | loop (Object{name=name, *)
(*  			     Type=tp, *)
(*  			     def=def}) = *)
(*  		  let val t=ff tp *)
(*  		      val d=ff def *)
(*  		  in fn a=>Object{name=name, *)
(*  				  Type=t a, *)
(*  				  def=d a} *)
(*  		  end *)
(*  	      | loop (ObjectInst{obj=obj,subst=f}) = *)
(*  		  let val obj=ff obj *)
(*  		  in fn a=>ObjectInst{obj=obj a,subst=f} *)
(*  		  end *)
(*  	      | loop (PatternFormal{name=name, *)
(*  				    Type=tp, *)
(*  				    extract=extract}) = *)
(*  		  let val t=ff tp *)
(*  		  in fn a=>PatternFormal{name=name, *)
(*  					 Type=t a, *)
(*  					 extract=extract} *)
(*  		  end *)
(*  	      | loop (TypeConstr{name=name, Type=tp}) = *)
(*  		  let val t=ff tp *)
(*  		  in fn a=>TypeConstr{name=name, Type=t a} *)
(*  		  end *)
(*  	      | loop (RecordField{name=name, Type=tp}) = *)
(*  		  let val t=ff tp *)
(*  		  in fn a=>RecordField{name=name, Type=t a} *)
(*  		  end *)
(*  	      | loop (StaticFormalConst{name=name,Type=def}) = *)
(*  		  let val d=ff def *)
(*  		  in fn a=>StaticFormalConst{name=name,Type=d a} *)
(*  		  end *)
(*  	      | loop (Theorem(p,name,def)) =  *)
(*  		  let val d=ff def *)
(*  		  in fn a=>Theorem(p,name,d a) *)
(*  		  end *)
(*  	      | loop (PType(p,tp)) =  *)
(*  		  let val t=ff tp *)
(*  		  in fn a=>PType(p,t a) *)
(*  		  end *)
(*  	      | loop (Symmetric(p,tp)) =  *)
(*  		  let val t=ff tp *)
(*  		  in fn a=>Symmetric(p,t a) *)
(*  		  end *)
(*  	      | loop (Finite(p,tp)) =  *)
(*  		  let val t=ff tp *)
(*  		  in fn a=>Finite(p,t a) *)
(*  		  end *)
(*  	      | loop (BeginEnd(p,lst)) = *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=>BeginEnd(p,l a) *)
(*  		  end *)
(*  	      | loop (Sync2(p,m1,m2)) = *)
(*  		  let val m1=ff m1 *)
(*  		      val m2=ff m2 *)
(*  		  in fn a=>Sync2(p, m1 a, m2 a) *)
(*  		  end *)
(*  	      | loop (Async2(p,m1,m2)) = *)
(*  		  let val m1=ff m1 *)
(*  		      val m2=ff m2 *)
(*  		  in fn a=>Async2(p, m1 a, m2 a) *)
(*  		  end *)
(*  	      | loop (Sync(p,lst,m)) = *)
(*  		  let val l=fflist lst *)
(*  		      val m=ff m *)
(*  		  in fn a=>Sync(p, l a, m a) *)
(*  		  end *)
(*  	      | loop (Async(p,lst,m)) =  *)
(*  		  let val l=fflist lst *)
(*  		      val m=ff m *)
(*  		  in fn a=>Async(p, l a, m a) *)
(*  		  end *)
(*  	      | loop (TypeDecl(p,name,parms,def)) = *)
(*  		  let val d=ff def *)
(*  		  in fn a=>TypeDecl(p,name,parms,d a) *)
(*  		  end *)
(*  	      | loop (DatatypeDecl(p,name,parms,def)) = *)
(*  		  let val d=ff def *)
(*  		  in fn a=> DatatypeDecl(p,name,parms,d a) *)
(*  		  end *)
(*  	      | loop (Var(p,lst)) = *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=>Var(p,l a) *)
(*  		  end *)
(*  	      | loop (StateVarDecl(p,lst)) = *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=>StateVarDecl(p,l a) *)
(*  		  end *)
(*  	      | loop (Val(p,name,def)) = *)
(*  		  let val d=ff def *)
(*  		  in fn a=>Val(p,name,d a) *)
(*  		  end *)
(*  	      | loop (FunCases(p,lst)) = *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=> FunCases(p,l a) *)
(*  		  end *)
(*  	      | loop (FunCase(p,name,pats,tp,def)) = *)
(*  		  let val pats=fflist pats *)
(*  		      val t=(case tp of *)
(*  				 NONE => NONE *)
(*  			       | SOME(t) => SOME(ff t)) *)
(*  		      val d=ff def *)
(*  		  in fn a=>FunCase(p,name,pats a, *)
(*  				   (case t of *)
(*  					NONE => NONE *)
(*  				      | SOME(x) => SOME(x a)), d a) *)
(*  		  end *)
(*  	      | loop (Open(p,m)) = *)
(*  		  let val m=ff m *)
(*  		  in fn a=>Open(p,m a) *)
(*  		  end *)
(*  	      | loop (Export(p,lst)) = *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=>Export(p,l a) *)
(*  		  end *)
(*  	      | loop (Models(p,x,y)) = *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>Models(p,x a,y a) *)
(*  		  end *)
(*  	      | loop (Refines(p,x,y)) =  *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>Refines(p,x a,y a) *)
(*  		  end *)
(*  	      | loop (NormalAsst(p,name,e)) = *)
(*  		  let val e=ff e *)
(*  		  in fn a=>NormalAsst(p,name,e a) *)
(*  		  end *)
(*  	      | loop (InitAsst(p,name,e)) = *)
(*  		  let val e=ff e *)
(*  		  in fn a=>InitAsst(p,name,e a) *)
(*  		  end *)
(*  	      | loop (NextAsst(p,name,e)) = *)
(*  		  let val e=ff e *)
(*  		  in fn a=> NextAsst(p,name,e a) *)
(*  		  end *)
(*  	      | loop (IfAsst(p,lst1,lst2)) = *)
(*  		  let val l1=fflist lst1 *)
(*  		      val l2=fflist lst2 *)
(*  		  in fn a=>IfAsst(p,l1 a,l2 a) *)
(*  		  end *)
(*  	      | loop (CondAsst(p,x,lst)) = *)
(*  		  let val x=ff x *)
(*  		      val l=fflist lst *)
(*  		  in fn a=>CondAsst(p, x a, l a) *)
(*  		  end *)
(*  	      | loop (LetAsst(p,defs,lst)) = *)
(*  		  let val d=fflist defs *)
(*  		      val l=fflist lst *)
(*  		  in fn a=>LetAsst(p,d a,l a) *)
(*  		  end *)
(*  	      | loop (CaseAsst(p,e,lst)) = *)
(*  		  let val e=ff e *)
(*  		      val l=fflist lst *)
(*  		  in fn a=>CaseAsst(p,e a,l a) *)
(*  		  end *)
(*  	      | loop (ChoiceAsst(p,x,lst)) = *)
(*  		  let val x=ff x *)
(*  		      val l=fflist lst *)
(*  		  in fn a=>ChoiceAsst(p,x a,l a) *)
(*  		  end *)
(*  	      | loop (ChooseAsst(p,lst)) = *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=>ChooseAsst(p,l a) *)
(*  		  end *)
(*  	      | loop (IfExpr(p,lst,x)) = *)
(*  		  let val l=fflist lst *)
(*  		      val x=ff x *)
(*  		  in fn a=>IfExpr(p, l a, x a) *)
(*  		  end *)
(*  	      | loop (CondExpr(p,x,y)) = *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>CondExpr(p,x a,y a) *)
(*  		  end *)
(*  	      | loop (LetExpr(p,defs,e)) = *)
(*  		  let val d=fflist defs *)
(*  		      val e=ff e *)
(*  		  in fn a=>LetExpr(p,d a, e a) *)
(*  		  end *)
(*  	      | loop (CaseExpr(p,e,lst)) = *)
(*  		  let val e=ff e *)
(*  		      val l=fflist lst *)
(*  		  in fn a=>CaseExpr(p, e a, l a) *)
(*  		  end *)
(*  	      | loop (ChoiceExpr(p,x,y)) = *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>ChoiceExpr(p, x a, y a) *)
(*  		  end *)
(*  	      | loop (ChooseExpr(p,x,y)) = *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=> ChooseExpr(p, x a, y a) *)
(*  		  end *)
(*  	      | loop (WithExpr(p,x,lst)) = *)
(*  		  let val x=ff x *)
(*  		      val l=fflist lst *)
(*  		  in fn a=>WithExpr(p, x a, l a) *)
(*  		  end *)
(*  	      | loop (WithAsst(p,name,y)) = *)
(*  		  let val y=ff y *)
(*  		  in fn a=>WithAsst(p,name, y a) *)
(*  		  end *)
(*  	      | loop (RecordExpr(p,lst)) = *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=>RecordExpr(p,l a) *)
(*  		  end *)
(*  	      | loop (TupleExpr(p,lst)) = *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=>TupleExpr(p,l a) *)
(*  		  end *)
(*  	      | loop (NondetExpr(p,lst)) = *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=> NondetExpr(p,l a) *)
(*  		  end *)
(*  	      | loop (RecordAsst(p,x,y)) = *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=> RecordAsst(p,x a, y a) *)
(*  		  end *)
(*  	      | loop (ModuleInst(p,m,lst)) = *)
(*  		  let val m=ff m *)
(*  		      val l=fflist lst *)
(*  		  in fn a=>ModuleInst(p, m a, l a) *)
(*  		  end *)
(*  	      | loop (Dot(p,x,y)) =  *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>Dot(p,x a, y a) *)
(*  		  end *)
(*  	      | loop (Type(p,tp)) =  *)
(*  		  let val t=ff tp *)
(*  		  in fn a=>Type(p,t a) *)
(*  		  end *)
(*  	      | loop (Datatype(p,tp)) =  *)
(*  		  let val t=ff tp *)
(*  		  in fn a=>Datatype(p,t a) *)
(*  		  end *)
(*  	      | loop (Forall(p,lst,x)) = *)
(*  		  let val l=fflist lst *)
(*  		      val x=ff x *)
(*  		  in fn a=>Forall(p,l a, x a) *)
(*  		  end *)
(*  	      | loop (Exists(p,lst,x)) =  *)
(*  		  let val l=fflist lst *)
(*  		      val x=ff x *)
(*  		  in fn a=>Exists(p,l a, x a) *)
(*  		  end *)
(*  	      | loop (RangeType(p,x,y)) =   *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>RangeType(p,x a, y a) *)
(*  		  end *)
(*  	      | loop (EnumType(p,lst)) =  *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=>EnumType(p,l a) *)
(*  		  end *)
(*  	      | loop (Of(p,name,tp)) = *)
(*  		  let val t=ff tp *)
(*  		  in fn a=>Of(p,name,t a) *)
(*  		  end *)
(*  	      | loop (TypedVars(p,lst,tp)) = *)
(*  		  let val l=fflist lst *)
(*  		      val t=(case tp of *)
(*  				 NONE => NONE *)
(*  			       | SOME(t) => SOME(ff t)) *)
(*  		  in fn a=>TypedVars(p,l a,(case t of *)
(*  						NONE => NONE *)
(*  					      | SOME(t) => SOME(t a))) *)
(*  		  end *)
(*  	      | loop (Appl(p,x,y)) = *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>Appl(p,x a, y a) *)
(*  		  end *)
(*  	      | loop (TypedExpr(p,x,y)) =  *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>TypedExpr(p,x a, y a) *)
(*  		  end *)
(*  	      | loop (TypedPattern(p,x,y)) =  *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>TypedPattern(p,x a, y a) *)
(*  		  end *)
(*  	      | loop (TuplePattern(p,lst)) =  *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=>TuplePattern(p,l a) *)
(*  		  end *)
(*  	      | loop (RecordPattern(p,lst)) =  *)
(*  		  let val l=fflist lst *)
(*  		  in fn a=>RecordPattern(p,l a) *)
(*  		  end *)
(*  	      | loop (ExtractAppl(x,y)) =  *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>ExtractAppl(x a, y a) *)
(*  		  end *)
(*  	      | loop (ExtractTuple(n,x)) =  *)
(*  		  let val x=ff x *)
(*  		  in fn a=>ExtractTuple(n,x a) *)
(*  		  end *)
(*  	      | loop (ExtractRecord(x,y)) =  *)
(*  		  let val x=ff x *)
(*  		      val y=ff y *)
(*  		  in fn a=>ExtractRecord(x a, y a) *)
(*  		  end *)
(*  	      | loop (Group(p,x)) =  *)
(*  		  let val x=ff x *)
(*  		  in fn a=>Group(p, x a) *)
(*  		  end *)
(*  	      | loop x = (fn a=>x) *)
(*  	in loop expr *)
(*  	end *)
  end
