(* This implementation should be more efficient, but it may not be as
   portable as the slow version below.  There are also destructive
   operations. *)

functor HashFun () : HASH =
  struct
    type ('a, 'b) Hash = { hash: ('a * 'b) list ref vector,
			   size: int,
			   F: 'a -> int,
			   eq: 'a * 'a -> bool }

    (* Somewhat arbitrary.  Makes it easy to implement F. *)
    val defaultSize = 1023

    (* The "almost perfect" hash function.  Converts a string into an
       integer from 0 to 255. *)
    fun F x =
	let val w = map(Word.fromInt o ord)(explode x)
	    val bytes = (Word.wordSize + 1) div 8
	    val size = Word.fromInt(List.length w)
	    fun fold _ acc [] = [acc]
	      | fold 1 acc lst = acc::(fold bytes 0wx0 lst)
	      | fold n acc (x::tl) = fold(n-1)(Word.orb(Word.<<(acc,0wx8),x)) tl
	    val z = fold bytes 0wx0 (size::w)
	in abs(Word.toInt(List.foldr(fn(x,y) => Word.xorb(x,y)) 0wx0 z))
	end

    fun makeHash eq = 
	 { hash=Vector.tabulate(1,(fn x => ref [])),
	   eq=eq,
	   F=fn x => 0,
	   size=1 }
    (* Ignore the size for now *)
    fun makeHashCustom(eq,size,F) = makeHash(eq)
    fun makeHashDefault(eq,toString) =
	{ hash=Vector.tabulate(defaultSize,(fn x => ref [])),
	   eq=eq,
	   F = (fn x => x mod defaultSize) o F o toString,
	   size=defaultSize }

    (* Make a clean copy of the hash, so any destructive operations wouldn't
     * change the original one *)
    fun copyHash{hash=hash,size=size,F=F,eq=eq} =
	{hash=Vector.map(fn x => ref(!x)) hash,
	 size=size,F=F,eq=eq}

    fun findHash({hash=hash,size=size,F=F,eq=eq},key) = 
	let fun strip (SOME(key,el)) = SOME el
	      | strip NONE = NONE
	in strip(List.find (fn (key1,_) => eq(key,key1))
		           (!(Vector.sub(hash,F(key)))))
	end

    fun insertHashDestructive(theHash,key,el) = 
	let val {hash=hash, eq=eq, F=F, size=_ } = theHash
	    val listref = Vector.sub(hash,F(key))
	in  (listref :=
	      ((key,el)::(List.filter(fn(key1,_)=>not(eq(key1,key)))
			             (!listref)));
	     theHash)
	end

    fun removeHashDestructive(theHash,key) = 
	let val {hash=hash, eq=eq, F=F, size=_ } = theHash
	    val listref = Vector.sub(hash,F(key))
	in  (listref := List.filter(fn(key1,_)=>not(eq(key1,key))) (!listref);
	     theHash)
	end

    fun insertHash(theHash,key,el) = 
	  insertHashDestructive(copyHash(theHash),key,el)

    fun removeHash(theHash,key) = 
	  removeHashDestructive(copyHash(theHash),key)

    fun wipeHash{hash=hash,size=size,F=F,eq=eq} =
	{ hash=Vector.tabulate(defaultSize,(fn x => ref [])),
	  eq=eq,
	  F=F,
	  size=size }

    fun hash2any (pkey: 'a -> 'c) (pval: 'b -> 'd)
	         ({hash=hash,...}: ('a,'b) Hash) = 
	let fun loop x = List.map(fn(key,el)=>(pkey key, pval el))(!x)
	    val tmp = Vector.map loop hash
	in Vector.foldr(op @) [] tmp
	end

    fun stats({hash=hash, ...}: ('a,'b) Hash) =
	Vector.foldr(fn(x,y) => (List.length(!x))::y) [] hash
  end


(* This hash is not supposed to be very efficient, and to be used only
   for syntax manipulation purposes. For instance, checking the program
   consistency and generating counterexample traces. *)

(* functor HashFun () : HASH =
  struct
    type ('a, 'b) Hash = { hash: ('a * 'b) list,
			   eq: 'a * 'a -> bool }

    fun makeHash eq = { hash=[], eq=eq }
    fun findHash({hash=hash,eq=eq},key) = 
	let fun strip (SOME(key,el)) = SOME el
	      | strip NONE = NONE
	in strip(List.find (fn (key1,el) => eq(key,key1)) hash)
	end
    fun insertHash({hash=hash,eq=eq},key,el) = 
	 { eq=eq,
	   hash= (key,el)::(List.filter 
			      (fn (key1,el) => not(eq(key1,key))) hash) }
    fun wipeHash{hash=hash,eq=eq} = {hash=[],eq=eq}

    fun hash2any (pkey: 'a -> 'c) (pval: 'b -> 'd) 
	         ({hash=hash,...}: ('a,'b) Hash) = 
	let fun loop [] = []
	      | loop ((key,el)::tl) = (pkey key, pval el)::(loop tl)
	in loop hash
	end
  end
*)
