functor InteractFun(structure Version: VERSION): INTERACT =
  struct
    structure Hash: HASH = HashFun()

    (* Command line options *)
    structure Options = OptionsFun()

    (* Parsing, reporting errors, etc. *)
    structure Interface
	= InterfaceFun(structure Options = Options)


    structure ProversJoin = 
	ProversJoinFun(structure Hash = Hash
		       structure Interface = Interface)

    open SympBug
    open Str
    open Options
    open Interface
    open Hash
    open ProversJoin
    open Version

    type SympState = unit

    local 
	fun initState() = ()
	val state = ref(initState())
    in
	fun conv ff args = ff(getOptions()) args
	val printError = conv printError 
	val printOut = conv printOut
	val printOutCommand = conv printOutCommand
	val verb = conv verb
	val verbStr = conv verbStr
	val lazyVerbDebug = conv lazyVerbDebug
	val verbDebug = conv verbDebug
	val lazyVerbDebugStr = conv lazyVerbDebugStr

	fun initSymp options =
	    (setOptions options;
	     initProvers())
	     

	(* Resets the entire tool.  Not implemented yet. *)
	fun ResetSympState() = ()
	    
	fun executeCommand UInull = verb("Null command")
	  | executeCommand UIexit = 
	      (verb("Exiting...");
	       exitAllProvers (getOptions()) (fn()=>(verb "Exiting...done";raise SympExit)))
	  | executeCommand UIyes = verb("Yes!..")
	  | executeCommand UIno = verb("No!..")
	  | executeCommand (UIcommand("typecheck",[UIstring file, UIstring tp])) =
	      typeCheckFile (SOME(str2ProverType tp)) (getOptions()) file
	  | executeCommand (UIcommand("typecheck",[UIstring file])) =
	      typeCheckFile NONE (getOptions()) file
	  | executeCommand (c as UIcommand("typecheck",lst)) = raise SympError
	      ("executeCommand: wrong arguments to typecheck:\n  "
	       ^(UI2string c)
	       ^"\n format: (typecheck \"fileName\"  [ \"proverType\" ])")
	  | executeCommand (UIcommand("version",_)) = printOut "version" version

	  (* `prover' commands always come from a particular prover
	     interface (e.g. prover emacs buffer), and must have the
	     prover ID and the command in the list of arguments *)

	  | executeCommand (UIcommand("prover",(UIstring id)::(UIstring command)::lst)) =
	      ProverCommand (getOptions()) (SOME id, command, lst)
	  | executeCommand (c as UIcommand("prover",lst)) = raise SympError
	      ("executeCommand: wrong arguments to `prover' command:\n  "
	       ^(UI2string c)
	       ^"\n format: (prover \"proverID\" \"command\" [ args ])")
	  | executeCommand (c as UIcommand("provetheorem",args)) =
	      let val options = getOptions()
		  val (thm, name, strName) =
		       (case args of
			    [UIstring thm] => raise SympError
			     ("no theorem name in `provetheorem' command:\n  "
			      ^(UI2string c)
			      ^"\n format: (provetheorem \"thm_spec\" [ \"new_name\" ])")
			  | [UIstring thm, UIstring str] => (thm, str, str)
			  | _ => raise SympError
			     ("executeCommand: wrong arguments to `provetheorem' command:\n  "
			      ^(UI2string c)
			      ^"\n format: (provetheorem \"thm_spec\" [ \"new_name\" ])"))
		  val proverID = proveTheorem options name thm
	      in 
		  (printOutCommand ("newprover",
				    [UIstring proverID, UIstring strName]);
		   printOutCommand ("proverstatus",
				    (UIstring proverID)::(List.map(fn x=>UIstring x)
							    (getProverStatus proverID)));
		   printCurrentSubgoal options proverID)
	      end
	  | executeCommand (UIcommand("userchoice",[UInumber id, UIstring choice])) =
	      acceptUserChoice(id, choice)
	  | executeCommand (c as UIcommand("userchoice",_)) = raise SympError
	      ("executeCommand: wrong arguments to `userchoice' command:\n  "
	       ^(UI2string c))
	  | executeCommand (UIcommand("cancelchoice",[UInumber id])) =
	      cancelUserChoice id
	  | executeCommand (c as UIcommand("cancelchoice",_)) = raise SympError
	      ("executeCommand: wrong arguments to `cancelchoice' command:\n  "
	       ^(UI2string c))
	  | executeCommand (C as UIcommand(c,args)) = raise SympError
	      ("Unknown command: "^(UI2string C))


	(* Wrapper for `executeCommand' to catch all the exceptions
	   and return the exit status `true' (success) or `false'
	   (exception caught).  The only exceptions we pass through is
	   `SympExit' (clean exit) and `SympError' (fatal error),
	   which are caught on the top level. *)

	fun tryExecute f = (f(); true)
	    handle SympError str =>
		     (printError "general" str; false)
		 | SympBug str =>
		     (printError "bug" ("\nOops, internal error!  "
					^"Please report this to sergey.berezin@cs.cmu.edu:\n\n"
					^str^"\n\nInclude your program and this output.\n"
					^"The internal SyMP state has been reset.");
		      (* lazyVerbDebug "" (typeCheckStat2string); *)
		      (* Since this is a nasty event, reset all the state and lose
		         all the typechecked/proved status info.  Play safe. *)
		      ResetSympState();
		      false)
		 | ParseError str => (printError "parse" str;
				      false)
		 | TypeError str => 
		     (printError "type" str;
		      (* lazyVerbDebug "" (typeCheckStat2string); *)
		      false)
		 | SequentError str => 
		     (printError "prover" str;
		      false)
		 | ProverError str => 
		     (printError "prover" str;
		      false)
		 | ProofTreeError str =>
		     (printError "prover" str;
		      false)
		 | SympExit => raise SympExit
		 (*  | _ => raise SympError("\nUnknow fatal error. Exiting.\n") *)

	fun tryExecuteCommand command = tryExecute(fn()=>executeCommand command)

	(* Same for a list of commands *)
	fun tryExecuteList [] = true
	  | tryExecuteList (c::lst) =
	    if tryExecuteCommand c then tryExecuteList lst
	    else false

	(* Read the user's command and execute it in an endless loop *)
	fun readExecLoop () = 
	    let val command = ReadUserCommand (getOptions()) ()
		handle SympError str => (printError ("\nError: "^str^"\n"); UInull)
		     | SympExit => (verb "Symp is exiting";
				    raise SympExit)
		     | SympExitError => (printError ("\nAbnormal exit\n");
					 raise SympExitError)
		     | _ => raise SympBug("readExecLoop: unknown exception caught")
	    in 
		(tryExecuteCommand command; readExecLoop())
	    end
	


    end
	

  end (* struct *)
