(* This is an abstract interface to a generic proof system.  Each
   implementation is specific to the input laguage, the sequent
   representation, the proof system, etc..  The implementation is also
   a bridge from a custom representation of each language and proof
   system to the uniform API used by the rest of the prover. *)

signature PROOF_SYSTEM =
  sig
    structure Sequent: SEQUENT
    type Sequent = Sequent.Sequent
    type InferenceRule = Sequent.InferenceRule
    type InferenceRuleResult = Sequent.InferenceRuleResult
    type InferenceRuleArgument = Sequent.InferenceRuleArgument
    type Substitution = Sequent.Substitution
    type ProgramContext = Sequent.ProgramContext

    (* Must be run once at the program startup to initialize the proof system *)
    val initProofSystem: unit -> unit

    val findRuleByName: string -> InferenceRule option

    (* Try to apply the rule and get a list of new sequents as
       premisses.  NONE means the rule doesn't apply.  May take a
       (partial) substitution obtained by the matchRule funciton.
       The first argument is a printing function to send messages to
       the user while applying the rule. *)
    val applyRule: (string -> unit) -> (Sequent * ProgramContext)
	        -> InferenceRule * (InferenceRuleArgument list) * (Substitution option)
	        -> InferenceRuleResult

    (* Each rule has a unique string identifier *)
    val getRuleName: InferenceRule -> string
    (* Help on the rule *)
    val getRuleHelp: InferenceRule -> string
    (* Add new inference rule, possibly overriding an existing one with the same name *)
    val addNewRule: InferenceRule -> unit
    (* Get the list of all the rules in the system *)
    val getAllRules: unit -> InferenceRule list
  end
