(* Implementation of a generic prover module *)
functor ProverFun(structure Interface: INTERFACE
		  structure ProofSystem: PROOF_SYSTEM
		  structure Hash: HASH
		  sharing Interface = ProofSystem.Sequent.Interface): PROVER =
  struct
    structure Interface = Interface
    structure ProofSystem = ProofSystem

    structure ProofTree = 
	ProofTreeFun(structure ProofSystem = ProofSystem
		     structure Hash = Hash
		     structure Interface = Interface)

    open SympBug
    open Str
    open Interface
    open ProofSystem
    open Sequent
    open SequentCommon
    open InputLanguage
    open Options
    open ProofTree

    type Prover = { id: string, (* unique ID for each prover *)
		    (* flag whether the proof has changed.
		       It gets reset when the proof is saved. *)
		    changed: bool ref,
		    context: ProgramContext,
		    proofrun: ProofRun }

    type ProverCommand = { name: string,
			   args: UserInputArgument list,
			   subgoal: Subgoal }

    (* Create a new prover object associated with a certain theorem
       and having an optional string ID (name) for the proof.  Don't
       mix it up with the prover ID, which is always unique can only
       be generated automatically when creating a new prover. *)

    local val proverID = ref 0
    in 
	fun makeProver name (obj, context) = 
	    let val seq = makeSequent context obj
		val run = makeProofRun(makeProofTree name)
		val tree = ProofRun2ProofTree run
		val prover = ({ id = Int.toString(!proverID),
			        changed = ref false,
				context = context,
			        proofrun = run }: Prover)
	    in (addSequent run (getRoot tree, seq);
		proverID := (!proverID) + 1;
		prover)
	    end
    end

    fun getProverID({id=id,...}: Prover) = id
    fun getProverChanged({changed=ref x,...}: Prover) = x
    fun setProverChanged({changed=xRef,...}: Prover) x = (xRef := x)
    fun getProverContext({context=c,...}: Prover) = c
    fun getCurrentProofTree({proofrun=pr,...}: Prover) = ProofRun2ProofTree pr
    fun getCurrentProofRun({proofrun=pr,...}: Prover) = pr

    (* Translate `args' into the `InferenceRuleArgument list' according
       to the `specs' of an inference rule arguments. 

       The way the args are translated: by default, we match the
       argument with the spec in the order they occur in the lists.
       If the current argument is a list, and the first element is a
       string and the second is "=", then it is a *named* argument,
       and the rest of the list is the value.  E.g., ["Gamma", "=", 1,
       2, 3] assigns "Gamma" the value [1,2,3].  If "Gamma" takes only
       scalar values, then only the first element of the list is
       taken. *)

    fun UIargs2IRargs(args: UserInputArgument list,
		      specs: InferenceRuleArgumentSpec list) =

	(* The war plan: take each argument in the list and try to
	   figure out the corresponding argument name.  Then
	   "match" the value, and construct the final IRArgValue.

	   Top level loop: (loop inorder args specs acc)
	   When inorder=true, we haven't seen any named arguments yet;
	   and if the argument is unnamed, pick the first spec and match it.
	   If inorder=false, the next argument MUST be named.
	   Each matched spec is removed from the `specs' list, not to be matched again.
	   All matched args are placed in `acc' list. *)

	let fun findSpecName specs name = 
	          List.find (fn s=> (IRArgSpecName s) = name) specs
	    (* match the argument with the spec and return `InferenceRuleArgument' *)
	    fun match(arg,spec) = 
		let val name = IRArgSpecName spec
		    val tp = IRArgSpecType spec
		    (* The actual translation of UIArg to IRArg *)
		    fun ff (UIstring str, ProverArgStringType) = ProverArgString str
		      | ff (UInumber n, ProverArgStringType) =
			  ProverArgString((if n >= 0 then "" else "-")^(Int.toString(abs n)))
		      | ff (UInumber n, ProverArgNumberType) = ProverArgNumber n
		      | ff (UIstring str, ProverArgSubgoalType) = ProverArgSubgoal str
		      | ff (arg, ProverArgFormulaNumType) = 
			  ProverArgFormulaNum(UIArg2FormulaIndex arg)
		      | ff (UIstring obj, ProverArgObjectType) =
			  ProverArgObject obj
		      | ff (UIlist lst, ProverArgListType tp) =
			  ProverArgList(List.map (fn arg => ff(arg, tp)) lst)
		      | ff (arg, tp) = raise ProverError
			  ("The argument has the wrong type: "^(UIArg2string arg)
			   ^"\n  Expected: "^(InfRuleArgType2string tp))
		in
		    { name = name,
		      value = ff(arg,tp) }: InferenceRuleArgument
		end
	    (* Same as `match', only fish out the spec with the given
	       name.  Return the rest of the specs and the IR argument. *)
	    fun matchNamed(name, arg, specs) = 
		(case findSpecName specs name of
		     NONE => raise ProverError
			 ("Unknown or duplicate named argument: "
			  ^(UIArg2string(UIassoc(UIstring name, arg))))
		   | SOME s =>
		      let val newSpecs = List.filter(fn s=> (IRArgSpecName s) <> name) specs
		      in
			  (newSpecs, match(arg, s))
		      end)
	    fun loop _ [] _ acc = acc
	      | loop inorder (args as _::_) [] _ = raise ProverError
	         ("Unmatched arguments: ["
		  ^(strlist2str ", " (List.map UIArg2string args))
		  ^"]")
	      | loop _ ((UIassoc(UIstring name, arg))::args) specs acc =
		 let val (newSpecs, newArg) = matchNamed(name, arg, specs)
		 in
		     loop false args newSpecs (newArg::acc)
		 end
	      | loop true (arg::args) (spec::specs) acc =
		 loop true args specs ((match(arg,spec))::acc)
	      | loop false (arg::args) specs acc =
		 raise ProverError("All unnamed arguments must preceed named ones: "
				   ^(UIArg2string arg))
	in 
	    loop true args specs []
	end

    (* Tries to apply a rule to the subgoal, and if succeeds, installs
       the rule on the subgoal and returns the list of new generated
       subgoals; otherwise return NONE.  The current subgoal must have
       a sequent in the proof run.  The new subgoals are guaranteed to
       have sequents. *)
    fun applyRule prover (command: ProverCommand) = 
	let val debug = lazyVerbDebug(getOptions()) "applyRule"
	    val { name=name,
		  args=args,
		  subgoal=subgoal } = command
	    val rule = (case findRuleByName name of
			    SOME r => r
			  | NONE => raise ProverError
				("No such proof rule: "^name
				 ^"\nType `help' for more information about rules and commands"))
	    val specs = getRuleArgs rule
	    val IRargs = UIargs2IRargs(args, specs)
	    val id = getProverID prover
	    val cxt = getProverContext prover
	    val proofRun = getCurrentProofRun prover
	    val _ = debug(fn()=>"\napplyRule: original proof tree:\n"
			  ^(ProofRun2stringDebug proofRun)^"\n")
	    val proofTreeID = getProofTreeID(getCurrentProofTree prover)
	    val sequent = (case getSequent proofRun subgoal of
			       SOME s => s
			     | NONE => raise ProverError
				   ("This subgoal doesn't have a sequent"))
	    fun printFun str = 
		  printOutCommand (getOptions()) ("prover", [UIstring id, UIstring str])
	    val newSequents = ProofSystem.applyRule printFun (sequent, cxt) (rule, IRargs, NONE)
	    (* Create a new subgoal for each sequent *)
	    fun newSubgoal(seq, hints) = 
		let val subgoal = makeSubgoal(SOME proofTreeID)
		in 
		    (subgoal, seq, hints)
		end
	    val triplesOpt =
		  (case  newSequents of
			InferenceRuleSuccess l => 
			    (debug(fn()=>"applyRule: generating new subgoals...");
			     (SOME(List.map newSubgoal l))
			     before (debug(fn()=>"done\n")))
		      | _ => NONE)
	    val proofRule = (rule, args)
	    fun triple2string(sub, seq, hints) =
		"("^(subgoal2stringDebug sub)
		^", sequent=[\n"^(sequent2string seq)
		^"]"
		^(case hints of
		      SOME h => ", hints=\""^(hints2string h)^"\""
		    | NONE => "")
		^")"
	    fun ff triples =
		let val subgoals = List.map(fn (s,_,h)=>(s,h)) triples
		    val _ = debug(fn()=>"\napplyRule: triples:\n"
				  ^(strlist2str "\n" (List.map triple2string triples))
				  ^"\n")
		in (debug(fn()=>"applyRule: installing new subgoals into the proof...");
		    setProverChanged prover true;
		    (* First add the subgoals to the proof tree *)
		    addProofRule(subgoal, proofRule, subgoals);
		    debug(fn()=> "\napplyRule: new proof tree without new sequents:\n"
			  ^(ProofRun2stringDebug proofRun));
		    (* Then associate them with the sequents *)
		    List.app(fn (sub,seq,_)=>addSequent proofRun (sub, seq)) triples;
		    debug(fn()=> "\napplyRule: new proof tree:\n"
			  ^(ProofRun2stringDebug proofRun));
		    (* debug(fn()=>"done\n"); *)
		    subgoals)
		end
	in
	    Option.map ff triplesOpt
	end

    (* Run the proof (ProofTree) on the given subgoal.  The current
       prover's proof tree is updated dynamically, step-by-step, so if
       the proof ultimately fails, at least some successful portion of
       it will be installed.  The result is NONE if the subgoal is
       proven, otherwise it is the list of unproven subgoals for the
       current subtree. *)
    fun runProof prover (tree, subgoal) = NONE

    (* I/O functions. *)

    (* If the subgoal has a sequent, it is printed along with the
       subgoal info *)

    fun subgoal2str prover subgoal =
	let val subInfo = Str((ProofTree.subgoal2string subgoal)^":\n\n")
	    val seq = (case getSequent (getCurrentProofRun prover) subgoal of
			   NONE => Str "No sequent for this subgoal"
			 | SOME s => sequent2str s)
	in Conc(subInfo, seq)
	end

    fun subgoal2string prover subgoal = Str2string(subgoal2str prover subgoal)

  (* Install the proof from the proof file *)
    fun installProofFileAt subgoal fileName =
	  installProofTree true (subgoal, file2ProofTree fileName)

    fun installProofFile prover fileName =
	let val root = getRoot(getCurrentProofTree prover)
	in
	    installProofFileAt root fileName
	end

    fun saveProof prover fileName =
	let val tree = getCurrentProofTree prover
	in
	    ProofTree2file(tree, fileName)
	end

  end
