(* Some functions that can be useful in implementing a custom
   'Sequent' module. *)

functor SequentCommonFun(structure Interface: INTERFACE
			 structure SequentBase: SEQUENT_BASE
			 sharing Interface.Options = SequentBase.InputLanguage.Options)
    : SEQUENT_COMMON =
  struct

    structure SequentBase = SequentBase
    structure Interface = Interface
    structure InputLanguage = SequentBase.InputLanguage
    open SequentBase
    open Interface
    open InputLanguage

    (* The result of the application of an inference rule.  *)
    datatype InferenceRuleResult =
	(* The rule applies and new subgoals are generated *)
	InferenceRuleSuccess of (Sequent * (SequentHints option)) list
	(* The rule proves the sequent false *)
      | InferenceRuleDisproven
	(* The rule doesn't apply *)
      | InferenceRuleFailure

    (* Actual argument to the inference rules.
       Each argument has a name and a value. *)
    datatype InferenceRuleArgumentValue =
	ProverArgString of string
      | ProverArgNumber of int
      | ProverArgSubgoal of string
	(* When formula number is NONE, this means all/any formulas from a given part *)
      | ProverArgFormulaNum of SequentPart * (int option)
	(* An Object in the input language.  Represent it as a string
           and let the rule decide how to parse/typecheck it *)
      | ProverArgObject of string
      | ProverArgList of InferenceRuleArgumentValue list

    type InferenceRuleArgument = 
	{ name: string,
	  value: InferenceRuleArgumentValue }

    (* Formal argument to an inference rule.  Each argument has a
       name, a "type", and an optional default value, whose presence
       means the argument is optional. *)
    datatype InferenceRuleArgumentType =
	ProverArgUnknownType
      | ProverArgStringType
      | ProverArgNumberType
      | ProverArgSubgoalType
      | ProverArgFormulaNumType
      | ProverArgObjectType
      | ProverArgListType of InferenceRuleArgumentType

    fun getInfRuleArgType (ProverArgString _) = ProverArgStringType
      | getInfRuleArgType (ProverArgNumber _) = ProverArgNumberType
      | getInfRuleArgType (ProverArgSubgoal _) = ProverArgSubgoalType
      | getInfRuleArgType (ProverArgFormulaNum _) = ProverArgFormulaNumType
      | getInfRuleArgType (ProverArgObject _) = ProverArgObjectType
      | getInfRuleArgType (ProverArgList []) = ProverArgListType ProverArgUnknownType
      | getInfRuleArgType (ProverArgList(x::_)) = ProverArgListType(getInfRuleArgType x)

    fun InfRuleArgType2string ProverArgStringType = "string"
      | InfRuleArgType2string ProverArgNumberType = "integer"
      | InfRuleArgType2string ProverArgSubgoalType = "subgoal"
      | InfRuleArgType2string ProverArgFormulaNumType = "formulaNum"
      | InfRuleArgType2string ProverArgObjectType = "expression"
      | InfRuleArgType2string (ProverArgListType tp) =
	  (InfRuleArgType2string tp)^" list"
      | InfRuleArgType2string ProverArgUnknownType = "unknown"

    fun IRArgStringValue(ProverArgString str) = SOME str
      | IRArgStringValue _ = NONE
    fun IRArgNumberValue(ProverArgNumber n) = SOME n
      | IRArgNumberValue _ = NONE
    fun IRArgSubgoalValue(ProverArgSubgoal str) = SOME str
      | IRArgSubgoalValue _ = NONE
    fun IRArgFormulaNumValue(ProverArgFormulaNum x) = SOME x
      | IRArgFormulaNumValue _ = NONE
    fun IRArgObjectValue(ProverArgString str) = SOME str
      | IRArgObjectValue _ = NONE
    fun IRArgListValue(ProverArgList lst) = SOME lst
      | IRArgListValue _ = NONE

    fun findIRArg args name =
	  Option.map(fn {value=v,...}: InferenceRuleArgument => v)
	            (List.find(fn {name=n,...} => n=name) args)

    fun findIRArgAnyValue ff args name = Option.mapPartial ff (findIRArg args name)

    val findIRArgValue = findIRArgAnyValue(fn x=>SOME x)
    val findIRArgStringValue = findIRArgAnyValue IRArgStringValue
    val findIRArgNumberValue = findIRArgAnyValue IRArgNumberValue
    val findIRArgSubgoalValue = findIRArgAnyValue IRArgSubgoalValue
    val findIRArgFormulaNumValue = findIRArgAnyValue IRArgFormulaNumValue
    val findIRArgObjectValue = findIRArgAnyValue IRArgObjectValue
    val findIRArgListValue = findIRArgAnyValue IRArgListValue

  end
